"""Ingestor for netmiko_hp_comware_display_ip_routing_table."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface, vrf, routetableentry
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device
    vrf_name = log.details.get("vrf")

    # Get or create VRF
    vrf_o = None
    if vrf_name:
        vrf_o = vrf.get(name=vrf_name)
        if not vrf_o:
            data = {
                "name": vrf_name,
            }
            vrf_o = vrf.create(**data)

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/hp_comware/display_ip_routing-table # pylint: disable=line-too-long
        nexthop_if_name = item.get("nexthop_if")
        distance = int(item.get("distance")) if item.get("distance") else None
        metric = int(item.get("metric")) if item.get("metric") else None
        destination = (
            f"{item.get('network')}/{item.get('mask')}" if item.get("network") else None
        )
        protocol = utils.normalize_route_type(item.get("protocal"))
        nexthop_ip = item.get("nexthop_ip") if item.get("nexthop_ip") else None

        # Get or create interface
        nexthop_if_id = None
        if nexthop_if_name:
            nexthop_if_label = utils.normalize_interface_label(nexthop_if_name)
            nexthop_if_o = interface.get(device_id=device_o.id, label=nexthop_if_label)
            if not nexthop_if_o:
                interface_data = {
                    "name": nexthop_if_label,
                    "device_id": device_o.id,
                }
                nexthop_if_o = interface.create(**interface_data)
            nexthop_if_id = nexthop_if_o.id

        # Get or create route table entry
        routetableentry_o = routetableentry.get(
            device_id=device_o.id,
            destination=destination,
            distance=distance,
            metric=metric,
            protocol=protocol,
        )
        if not routetableentry_o:
            data = {
                "device_id": device_o.id,
                "destination": destination,
                "distance": distance,
                "metric": metric,
                "protocol": protocol,
                "nexthop_if_id": nexthop_if_id,
                "nexthop_ip": nexthop_ip,
                "vrf_id": vrf_o.id if vrf_o else None,
            }

            routetableentry.create(**data)

    # Update the log
    log.ingested = True
    log.save()



"""Schema validation for VLAN."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from ipam.models import VLAN
from ipam.choices import VLANStatusChoices

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate VLAN data."""
    return {
        "type": "object",
        "properties": {
            "vid": {
                "type": "integer",
            },
            "name": {
                "type": "string",
            },
            "status": {
                "type": "string",
                "enum": [key for key, value in VLANStatusChoices()],
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new VLAN objects."""
    schema = get_schema()
    schema["required"] = [
        "vid",
        "name",
        "status",
    ]
    return schema


def create(status="active", **kwargs):
    """Create an VLAN."""
    data = {
        **kwargs,
        "status": status,  # Default status is active
    }
    data = utils.delete_empty_keys(data)
    validate(data, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(VLAN, **data)
    return obj


def get(vid, name):
    """Return an VLAN."""
    obj = utils.object_get_or_none(VLAN, vid=vid, name=name)
    return obj


def get_list(**kwargs):
    """Get a list of VLAN objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(VLAN, **kwargs)
    return result



"""Sidebar navigation buttons."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from extras.plugins import PluginMenuButton, PluginMenuItem, PluginMenu
from utilities.choices import ButtonColorChoices

credential_buttons = [
    PluginMenuButton(
        link="plugins:netdoc:credential_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        color=ButtonColorChoices.GREEN,
    ),
    PluginMenuButton(
        link="plugins:netdoc:credential_import",
        title="Import",
        icon_class="mdi mdi-upload",
        color=ButtonColorChoices.CYAN,
    ),
]

diagram_buttons = [
    PluginMenuButton(
        link="plugins:netdoc:diagram_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        color=ButtonColorChoices.GREEN,
    ),
]

discoverable_buttons = [
    PluginMenuButton(
        link="plugins:netdoc:discoverable_add",
        title="Add",
        icon_class="mdi mdi-plus-thick",
        color=ButtonColorChoices.GREEN,
    ),
    PluginMenuButton(
        link="plugins:netdoc:discoverable_import",
        title="Import",
        icon_class="mdi mdi-upload",
        color=ButtonColorChoices.CYAN,
    ),
]

menu_discovery = (
    PluginMenuItem(
        link="plugins:netdoc:credential_list",
        link_text="Credentials",
        buttons=credential_buttons,
    ),
    PluginMenuItem(
        link="plugins:netdoc:discoverable_list",
        link_text="Discoverables",
        buttons=discoverable_buttons,
    ),
    PluginMenuItem(
        link="plugins:netdoc:discoverylog_list",
        link_text="Logs",
    ),
)

menu_tables = (
    PluginMenuItem(
        link="plugins:netdoc:arptableentry_list",
        link_text="ARP Table",
    ),
    PluginMenuItem(
        link="plugins:netdoc:macaddresstableentry_list",
        link_text="MAC Address Table",
    ),
    PluginMenuItem(
        link="plugins:netdoc:routingtable_list",
        link_text="Routing Table",
    ),
)

menu = PluginMenu(
    label="NetDoc",
    groups=(
        (
            "Diagrams",
            (
                PluginMenuItem(
                    link="plugins:netdoc:diagram_list",
                    link_text="Diagrams",
                    buttons=diagram_buttons,
                ),
            ),
        ),
        ("Discovery", menu_discovery),
        ("Tables", menu_tables),
    ),
)



"""Schema validation for Credential."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from netdoc.models import Credential

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate Credential data."""
    return {
        "type": "object",
        "properties": {
            "enable_password": {
                "type": "string",
            },
            "name": {
                "type": "string",
            },
            "password": {
                "type": "string",
            },
            "username": {
                "type": "string",
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Credential objects."""
    schema = get_schema()
    schema["required"] = [
        "name",
    ]
    return schema


def get(name):
    """Return a Credential given name."""
    obj = utils.object_get_or_none(Credential, name=name)
    return obj


def get_or_create(name, **kwargs):
    """Get or create a Credential."""
    created = False
    data = {
        **kwargs,
        "name": name,
    }
    data = utils.delete_empty_keys(data)
    validate(data, get_schema_create(), format_checker=FormatChecker())

    obj = utils.object_get_or_none(Credential, name=name)
    if not obj:
        obj = utils.object_create(Credential, name=name, **kwargs)
        created = True

    return obj, created


def get_list(**kwargs):
    """Get a list of Credential objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Credential, **kwargs)
    return result



# Generated by Django 4.1.6 on 2023-02-12 09:11

from django.db import migrations, models
import utilities.json


class Migration(migrations.Migration):
    dependencies = [
        ("netdoc", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="arptableentry",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
        migrations.AlterField(
            model_name="credential",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
        migrations.AlterField(
            model_name="discoverable",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
        migrations.AlterField(
            model_name="discoverylog",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
        migrations.AlterField(
            model_name="macaddresstableentry",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
        migrations.AlterField(
            model_name="routetableentry",
            name="custom_field_data",
            field=models.JSONField(
                blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder
            ),
        ),
    ]



"""Views, called by URLs."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import logging
import json

from django.db.models import Count
from django.contrib import messages
from django.shortcuts import redirect, render
from django.urls import reverse
from django.http import HttpResponse

from utilities.forms import ConfirmationForm
from utilities.htmx import is_htmx
from utilities.permissions import get_permission_for_model
from utilities.utils import get_viewname
from netbox.views import generic

from netdoc import models, tables, forms, filtersets, utils, topologies

#
# ARPEntry views
#


class ArpTableListView(generic.ObjectListView):
    """Summary view listing all ARP."""

    queryset = models.ArpTableEntry.objects.all().order_by(
        "interface__device__name", "interface__name", "ip_address"
    )
    table = tables.ArpTableEntryTable
    filterset = filtersets.ArpTableEntryFilterSet
    actions = []  # Read only table


class ArpTableView(generic.ObjectView):
    """Detailed ARP table entry view."""

    queryset = models.ArpTableEntry.objects.all()

    def get_extra_context(self, request, instance):
        """Get associated ARP and MAC Address tables."""
        arp_table_qs = models.ArpTableEntry.objects.filter(
            ip_address=str(instance.ip_address)
        )
        arp_table = tables.ArpTableEntryTable(arp_table_qs)
        arp_table.configure(request)

        macaddress_table_qs = models.MacAddressTableEntry.objects.filter(
            mac_address=instance.mac_address
        )
        macaddress_table = tables.MacAddressTableEntryTable(macaddress_table_qs)
        macaddress_table.configure(request)

        return {
            "arp_table": arp_table,
            "macaddress_table": macaddress_table,
        }


#
# Credential views
#


class CredentialListView(generic.ObjectListView):
    """Summary view listing all Credential objects."""

    queryset = models.Credential.objects.annotate(
        discoverables_count=Count("discoverables")
    ).order_by("name")
    table = tables.CredentialTable
    filterset = filtersets.CredentialFilterSet


class CredentialView(generic.ObjectView):
    """Detailed Credential view."""

    queryset = models.Credential.objects.all()

    def get_extra_context(self, request, instance):
        """Get associated Discoverable obhects."""
        table = tables.DiscoverableTableWOLogCount(instance.discoverables.all())
        table.configure(request)

        return {
            "discoverables_table": table,
        }


class CredentialEditView(generic.ObjectEditView):
    """Edit Credential view."""

    queryset = models.Credential.objects.all()
    form = forms.CredentialForm


class CredentialDeleteView(generic.ObjectDeleteView):
    """Delete Credential view."""

    queryset = models.Credential.objects.all()
    default_return_url = "plugins:netdoc:credential_list"


class CredentialBulkImportView(generic.BulkImportView):
    """Bulk import Credential view."""

    queryset = models.Credential.objects.all()
    model_form = forms.CredentialCSVForm
    table = tables.CredentialTable


class CredentialBulkEditView(generic.BulkEditView):
    """Bulk edit Credential view."""

    queryset = models.Credential.objects.all()
    table = tables.CredentialTable
    default_return_url = "plugins:netdoc:credential_list"
    filterset = filtersets.CredentialFilterSet


class CredentialBulkDeleteView(generic.BulkDeleteView):
    """Bulk delete Credential view."""

    queryset = models.Credential.objects.all()
    table = tables.CredentialTable
    default_return_url = "plugins:netdoc:credential_list"
    filterset = filtersets.CredentialFilterSet


#
# Diagram views
#


class DiagramListView(generic.ObjectListView):
    """Summary view listing all Diagram objects."""

    queryset = models.Diagram.objects.all().order_by("name")
    table = tables.DiagramTable


class DiagramView(generic.ObjectView):
    """Detailed Diagram view."""

    queryset = models.Diagram.objects.all()

    def get_extra_context(self, request, instance):
        """Get associated Diagram obhects from Interface queryset."""
        sites = list(instance.sites.all().values_list("id", flat=True))
        roles = list(instance.device_roles.all().values_list("id", flat=True))
        vrfs = list(instance.vrfs.all().values_list("id", flat=True))

        interface_qs = utils.get_diagram_interfaces(
            instance.mode,
            sites=sites,
            roles=roles,
            vrfs=vrfs,
            include_global_vrf=instance.include_global_vrf,
        )

        # Build and return the topology
        try:
            module = getattr(topologies, f"get_{instance.mode}_topology_data")
        except ModuleNotFoundError as exc:
            raise ModuleNotFoundError(
                f"Get topology function not found for {instance.mode}"
            ) from exc
        return {
            "topology_data": module(interface_qs, instance.details),
            "topology_details": instance.details,
        }


class DiagramEditView(generic.ObjectEditView):
    """Edit Diagram view."""

    queryset = models.Diagram.objects.all()
    form = forms.DiagramForm


class DiagramDeleteView(generic.ObjectDeleteView):
    """Delete Diagram view."""

    queryset = models.Diagram.objects.all()
    default_return_url = "plugins:netdoc:diagram_list"


class DiagramExportView(generic.ObjectDeleteView):
    """
    Export a single object.

    Called from:
    * DiagramView clicking on the Export button on a specific Diagram.
    """

    queryset = models.Diagram.objects.all()

    def get(self, request, *args, **kwargs):
        """Download the log."""
        instance = self.get_object(**kwargs)
        sites = list(instance.sites.all().values_list("id", flat=True))
        roles = list(instance.device_roles.all().values_list("id", flat=True))
        vrfs = list(instance.vrfs.all().values_list("id", flat=True))

        interface_qs = utils.get_diagram_interfaces(
            instance.mode,
            sites=sites,
            roles=roles,
            vrfs=vrfs,
            include_global_vrf=instance.include_global_vrf,
        )

        # Build and return the topology
        try:
            module = getattr(topologies, f"get_{instance.mode}_drawio_topology")
        except ModuleNotFoundError as exc:
            raise ModuleNotFoundError(
                f"Get DrawIO topology function not found for {instance.mode}"
            ) from exc
        drawio_xml = module(interface_qs, instance)

        response = HttpResponse(drawio_xml, content_type="application/xml")
        response["Content-Disposition"] = f"attachment; filename={instance.name}.drawio"
        return response


#
# Discoverable views
#


class DiscoverableListView(generic.ObjectListView):
    """Summary view listing all Discoverable objects."""

    queryset = models.Discoverable.objects.annotate(
        discoverylogs_count=Count("discoverylogs")
    ).order_by("device__name", "address")
    table = tables.DiscoverableTable
    actions = [
        "add",
        "import",
        "export",
        "discover",
        "bulk_edit",
        "bulk_delete",
        "bulk_discover",
    ]
    template_name = "netdoc/discoverable_list.html"
    filterset = filtersets.DiscoverableFilterSet


class DiscoverableView(generic.ObjectView):
    """Detailed Discoverable view."""

    queryset = models.Discoverable.objects.annotate(
        discoverylogs_count=Count("discoverylogs")
    )

    def get_extra_context(self, request, instance):
        """Get associated DiscoveryLog obhects."""
        table = tables.DiscoveryLogTable(
            instance.discoverylogs.all().order_by("-created")
        )
        table.configure(request)

        return {
            "discoverylogs_table": table,
        }


class DiscoverableEditView(generic.ObjectEditView):
    """Edit Discoverable view."""

    queryset = models.Discoverable.objects.all()
    form = forms.DiscoverableForm


class DiscoverableDeleteView(generic.ObjectDeleteView):
    """Delete Discoverable view."""

    queryset = models.Discoverable.objects.all()
    default_return_url = "plugins:netdoc:discoverable_list"


class DiscoverableBulkImportView(generic.BulkImportView):
    """Bulk import Discoverable view."""

    queryset = models.Discoverable.objects.all()
    model_form = forms.DiscoverableCSVForm
    table = tables.DiscoverableTable


class DiscoverableBulkEditView(generic.BulkEditView):
    """Bulk edit Discoverable view."""

    queryset = models.Discoverable.objects.all()
    table = tables.DiscoverableTable
    form = forms.DiscoverableBulkEditForm


class DiscoverableBulkDeleteView(generic.BulkDeleteView):
    """Bulk delete Discoverable view."""

    queryset = models.Discoverable.objects.all()
    table = tables.DiscoverableTable
    default_return_url = "plugins:netdoc:discoverable_list"


class DiscoverableDiscoverView(generic.ObjectDeleteView):
    """
    Discover a single object.

    Called from:
    * DiscoverableListView clicking on the Discovery button on a specific Discoverable row.
    * DiscoverableView clicking on the Discovery button.
    """

    queryset = models.Discoverable.objects.all()
    template_name = "netdoc/discoverable_discover.html"

    def get_required_permission(self):
        """Check permissions."""
        return get_permission_for_model(self.queryset.model, "change")

    #
    # Request handlers
    #

    def get(self, request, *args, **kwargs):
        """Return the confirmation page."""
        obj = self.get_object(**kwargs)
        form = ConfirmationForm(initial=request.GET)

        # If this is an HTMX request, return only the rendered deletion form as modal content
        if is_htmx(request):
            # Called from DiscoverableView
            viewname = get_viewname(self.queryset.model, action="discover")
            form_url = reverse(viewname, kwargs={"pk": obj.pk})
            return render(
                request,
                "netdoc/htmx/discover_form.html",
                {
                    "object": obj,
                    "object_type": self.queryset.model._meta.verbose_name,  # pylint: disable=protected-access
                    "form": form,
                    "form_url": form_url,
                    **self.get_extra_context(request, obj),
                },
            )

        # Called from DiscoverableViewList
        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )

    def post(self, request, *args, **kwargs):
        """Start the discovery on a single Discoverable."""
        logger = logging.getLogger("netbox.plugins.netdoc")
        obj = self.get_object(**kwargs)
        form = ConfirmationForm(request.POST)

        if form.is_valid():
            logger.debug("Form validation was successful")
            discoverables = [obj]

            # Starting discovery job on default queue (single Discoverable)
            msg = f"Starting discovery on {obj}"
            logger.info(msg)
            messages.success(request, msg)
            data = {
                "discoverables": discoverables,
            }
            utils.spawn_script("Discover", user=request.user, post_data=data)

            return_url = form.cleaned_data.get("return_url")
            if return_url and return_url.startswith("/"):
                return redirect(return_url)
            return redirect(self.get_return_url(request, obj))

        logger.debug("Form validation failed")

        return render(
            request,
            self.template_name,
            {
                "object": obj,
                "form": form,
                "return_url": self.get_return_url(request, obj),
                **self.get_extra_context(request, obj),
            },
        )


class DiscoverableBulkDiscoverView(generic.BulkDeleteView):
    """
    Disocver devices in bulk.

    Called from:
    * DiscoverableListView selecting Discoverable(s) and clicking on Disocver Selected button.
    """

    template_name = "netdoc/discoverable_bulk_discover.html"
    queryset = models.Discoverable.objects.prefetch_related("credential")
    filterset = None
    table = tables.DiscoverableTable
    default_return_url = "plugins:netdoc:discoverable_list"

    def get_required_permission(self):
        """Check permissions."""
        return get_permission_for_model(self.queryset.model, "change")

    def post(self, request, **kwargs):
        """Start the discovery."""
        logger = logging.getLogger("netbox.plugins.netdoc")
        model = self.queryset.model

        # Are we discovering *all* objects in the queryset or just a selected subset?
        if request.POST.get("_all"):
            queryset = model.objects.all()
            pk_list = queryset.only("pk").values_list("pk", flat=True)
        else:
            pk_list = [int(pk) for pk in request.POST.getlist("pk")]

        form_cls = self.get_form()

        if "_confirm" in request.POST:
            form = form_cls(request.POST)
            if form.is_valid():
                logger.debug("Form validation was successful")
                queryset = self.queryset.filter(pk__in=pk_list)
                discovery_count = queryset.count()

                # Starting discovery job on default queue (list of Discoverable)
                msg = f"Starting discovery on {discovery_count} {model._meta.verbose_name_plural}"  # pylint: disable=protected-access
                logger.info(msg)
                messages.success(request, msg)
                data = {
                    "discoverables": list(queryset),
                }
                utils.spawn_script("Discover", user=request.user, post_data=data)

                return redirect(self.get_return_url(request))

            logger.debug("Form validation failed")

        else:
            form = form_cls(
                initial={
                    "pk": pk_list,
                    "return_url": self.get_return_url(request),
                }
            )

        # Retrieve objects being deleted
        table = self.table(self.queryset.filter(pk__in=pk_list), orderable=False)
        if not table.rows:
            messages.warning(
                request,
                f"No {model._meta.verbose_name_plural} were selected for discovery.",  # pylint: disable=protected-access
            )
            return redirect(self.get_return_url(request))

        return render(
            request,
            self.template_name,
            {
                "model": model,
                "form": form,
                "table": table,
                "return_url": self.get_return_url(request),
                **self.get_extra_context(request),
            },
        )


#
# DiscoveryLog views
#


class DiscoveryLogListView(generic.ObjectListView):
    """Summary view listing all DiscoveryLog objects."""

    queryset = models.DiscoveryLog.objects.all().order_by("-created")
    table = tables.DiscoveryLogTable
    filterset = filtersets.DiscoveryLogFilterSet
    filterset_form = forms.DiscoveryLogListFilterForm
    actions = ["delete", "bulk_delete"]  # Read-only + delete + bulk-delete delete


class DiscoveryLogView(generic.ObjectView):
    """Detailed DiscoveryLog view."""

    queryset = models.DiscoveryLog.objects.all()


class DiscoveryLogDeleteView(generic.ObjectDeleteView):
    """Delete DiscoveryLog view."""

    queryset = models.DiscoveryLog.objects.all()
    default_return_url = "plugins:netdoc:discoverylog_list"


class DiscoveryLogExportView(generic.ObjectView):
    """
    Export a single object.

    Called from:
    * DiscoveryLogView clicking on the Discovery button on a specific Discoverable.
    """

    queryset = models.DiscoveryLog.objects.all()

    def get(self, request, **kwargs):
        """Download the log."""
        instance = self.get_object(**kwargs)
        data = utils.export_log(instance)
        response = HttpResponse(json.dumps(data), content_type="application/json")
        response["Content-Disposition"] = f"attachment; filename={instance.id}.json"
        return response


class DiscoveryLogBulkDeleteView(generic.BulkDeleteView):
    """Bulk delete DiscoveryLog view."""

    queryset = models.DiscoveryLog.objects.all()
    table = tables.DiscoveryLogTable
    default_return_url = "plugins:netdoc:discoverylog_list"


#
# MacAddressTableEntry views
#


class MacAddressTableListView(generic.ObjectListView):
    """Summary view listing all MAC address."""

    queryset = models.MacAddressTableEntry.objects.all().order_by(
        "interface__device__name", "interface__name", "mac_address"
    )
    table = tables.MacAddressTableEntryTable
    filterset = filtersets.MacAddressTableEntryFilterSet
    actions = []  # Read only table


class MacAddressTableView(generic.ObjectView):
    """Detailed MAC address entry view."""

    queryset = models.MacAddressTableEntry.objects.all()
    actions = []  # Read only table

    def get_extra_context(self, request, instance):
        """Get associated MAC Address tables."""
        arp_table_qs = models.ArpTableEntry.objects.filter(
            mac_address=str(instance.mac_address)
        )
        arp_table = tables.ArpTableEntryTable(arp_table_qs)
        arp_table.configure(request)

        macaddress_table_qs = models.MacAddressTableEntry.objects.filter(
            mac_address=instance.mac_address
        )
        macaddress_table = tables.MacAddressTableEntryTable(macaddress_table_qs)
        macaddress_table.configure(request)

        return {
            "arp_table": arp_table,
            "macaddress_table": macaddress_table,
        }


#
# RouteTableEntry view
#


class RouteTableEntryListView(generic.ObjectListView):
    """Summary view listing all routes."""

    queryset = models.RouteTableEntry.objects.all().order_by(
        "device__name", "destination", "protocol", "nexthop_if__name"
    )
    table = tables.RouteTableEntryTable
    filterset = filtersets.RouteTableEntryFilterSet
    actions = []  # Read only table



"""Ingestor for netmiko_cisco_ios_show_inventory."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_inventory # pylint: disable=line-too-long
        part_description = item.get("name")
        part_serial_number = item.get("sn")
        # part_id = item.get("pid")

        if "chassis" in part_description.lower():
            # Chassis Serial Number
            device_o.serial = part_serial_number
            device_o.save()
            break

    # Update the log
    log.ingested = True
    log.save()



"""Custom Inventory for Nornir."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from nornir.core.inventory import (
    Inventory,
    Host,
    Hosts,
    Group,
    Groups,
    ParentGroups,
    Defaults,
    ConnectionOptions,
)

from django.conf import settings

from netdoc import models

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("netdoc", {})


class AssetInventory:
    """Build Nornir AssetInventory.

    AssetInventory is an inventory plugin for Nornir that loads data from
    Discoverable table. Can be registered and used with:

    from nornir.core.plugins.inventory import InventoryPluginRegister
    from netdoc.nornir_inventory import AssetInventory
    from nornir import InitNornir

    InventoryPluginRegister.register("asset-inventory", AssetInventory)
    nr = InitNornir(
        runner={
            "plugin": "threaded",
            "options": {
                "num_workers": 100,
            },
        },
        inventory={
            "plugin": "asset-inventory",
        },
        logging={"enabled": False},
    )
    """

    def load(self) -> Inventory:
        """Load items from remote API."""
        defaults = Defaults()
        hosts = Hosts()
        groups = Groups()

        discoverables = models.Discoverable.objects.filter(discoverable=True)

        # Add "all" group
        groups["all"] = Group("all")

        # Load discoverable hosts
        for discoverable in discoverables:
            credential = discoverable.credential
            # Add hosts discoverable via Netmiko
            device_type = "_".join(discoverable.mode.split("_")[1:])
            data = {
                "site_id": discoverable.site.pk,
                "site": discoverable.site.slug,
            }

            host_key = discoverable.address
            host_groups = [device_type, f'site-{data["site"]}']

            # Create additional options
            extras = {}
            if credential.enable_password:
                extras["secret"] = credential.enable_password
            connection_options = {"netmiko": ConnectionOptions(extras=extras)}

            hosts[host_key] = Host(
                name=host_key,
                hostname=discoverable.address,
                username=credential.username,
                password=credential.password,
                port=22,
                platform=device_type,
                data=data,
                groups=ParentGroups(),
                connection_options=connection_options,
            )  # name is the key used in AggregatedResults, the form is: tenant:id:ip_address

            # Add groups
            for host_group in host_groups:
                if host_group not in dict(groups):
                    groups[host_group] = Group(host_group)
                hosts[host_key].groups.append(Group(host_group))

        return Inventory(hosts=hosts, groups=groups, defaults=defaults)



"""
Clean previous data, load lab-specific content, and test.

To make the process faster, the database is created once and resued:
    - /opt/netbox/venv/bin/python3 /opt/netbox/netbox/manage.py test netdoc --verbosity=2 --keepdb
"""

from os import walk, path
import re
import json
import yaml
from yaml.loader import SafeLoader

from django.test import TestCase

from dcim.models import (
    Device,
    DeviceRole,
    DeviceType,
    Manufacturer,
    Cable,
    CableTermination,
    Interface,
    CablePath,
    Site,
)
from ipam.models import IPAddress, Prefix, VRF, VLAN

from netdoc.models import (
    ArpTableEntry,
    MacAddressTableEntry,
    RouteTableEntry,
    Discoverable,
    Credential,
    DiscoveryLog,
)
from netdoc.schemas import (
    credential as credential_api,
    discoverable as discoverable_api,
    discoverylog as discoverylog_api,
)
from netdoc.utils import log_ingest

TEST_DIRECTORY = path.dirname(path.realpath(__file__))


def test_devices(test_o, expected_results):
    """Test Device given an expected_results dict."""
    # Test total Device objects
    device_qs = Device.objects.all()
    test_o.assertEquals(len(device_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        device_o = Device.objects.get(name=expected_result.get("name"))
        test_o.assertEquals(
            device_o.device_type.manufacturer.name, expected_result.get("manufacturer")
        )
        test_o.assertEquals(device_o.serial, expected_result.get("serial"))

        if device_o.site:
            test_o.assertEquals(device_o.site.name, expected_result.get("site"))
        else:
            test_o.assertIs(expected_result.get("site"), None)

        if device_o.primary_ip:
            test_o.assertEquals(
                str(device_o.primary_ip.address), expected_result.get("address")
            )
        else:
            test_o.assertIs(expected_result.get("address"), None)


def test_discoverables(test_o, expected_results):
    """Test Discoverable given an expected_results dict."""
    # Test total Discoverable objects
    discoverable_qs = Discoverable.objects.all()
    test_o.assertEquals(len(discoverable_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        discoverable_o = Discoverable.objects.get(
            address=expected_result.get("address")
        )
        test_o.assertEquals(discoverable_o.mode, expected_result.get("mode"))

        if discoverable_o.device:
            test_o.assertEquals(
                discoverable_o.device.name, expected_result.get("device")
            )
        else:
            test_o.assertIs(expected_result.get("device"), None)

        if discoverable_o.site:
            test_o.assertEquals(discoverable_o.site.name, expected_result.get("site"))
        else:
            test_o.assertIs(expected_result.get("site"), None)

        if discoverable_o.credential:
            test_o.assertEquals(
                discoverable_o.credential.name, expected_result.get("credential")
            )
        else:
            test_o.assertIs(expected_result.get("credential"), None)


def test_interfaces(test_o, expected_results):
    """Test Interface given an expected_results dict."""
    # Test total Interface objects
    ipaddress_qs = Interface.objects.all()
    test_o.assertEquals(
        len(ipaddress_qs),
        len(
            [
                interface_value
                for device_name in list(expected_results.values())
                for interface_value in device_name
            ]
        ),
    )

    # Test each device
    for device_name, interface_list in expected_results.items():
        # Test each interface
        for interface_value in interface_list:
            interface_o = Interface.objects.get(
                label=interface_value.get("label"), device__name=device_name
            )
            test_o.assertEquals(interface_o.type, interface_value.get("type"))
            test_o.assertEquals(interface_o.speed, interface_value.get("speed"))
            test_o.assertEquals(interface_o.duplex, interface_value.get("duplex"))
            test_o.assertEquals(interface_o.mtu, interface_value.get("mtu"))
            test_o.assertEquals(interface_o.enabled, interface_value.get("enabled"))
            test_o.assertEquals(
                interface_o.description, interface_value.get("description")
            )
            test_o.assertEquals(interface_o.mode, interface_value.get("mode"))

            if interface_value.get("name"):
                # Check Interface.name only if is not None on interfaces.yml
                test_o.assertEquals(interface_o.name, interface_value.get("name"))

            if interface_o.mac_address:
                test_o.assertEquals(
                    str(interface_o.mac_address), interface_value.get("mac_address")
                )
            else:
                test_o.assertIs(interface_value.get("mac_address"), None)

            if interface_o.vrf:
                test_o.assertEquals(interface_o.vrf.name, interface_value.get("vrf"))
            else:
                test_o.assertIs(interface_value.get("vrf"), None)

            if interface_o.parent:
                test_o.assertEquals(
                    interface_o.parent.label, interface_value.get("parent")
                )
            else:
                test_o.assertIs(interface_value.get("parent"), None)

            if interface_o.lag:
                test_o.assertEquals(interface_o.lag.label, interface_value.get("lag"))
            else:
                test_o.assertIs(interface_value.get("lag"), None)

            if interface_o.connected_endpoints:
                test_o.assertEquals(
                    interface_o.connected_endpoints[0].device.name,
                    interface_value.get("connected_device"),
                )
                test_o.assertEquals(
                    interface_o.connected_endpoints[0].label,
                    interface_value.get("connected_interface_label"),
                )
            else:
                test_o.assertIs(interface_value.get("connected_device"), None)
                test_o.assertIs(interface_value.get("connected_interface_label"), None)

            if interface_o.untagged_vlan:
                test_o.assertEquals(
                    interface_o.untagged_vlan.vid, interface_value.get("untagged_vlan")
                )
            else:
                test_o.assertIs(interface_value.get("untagged_vlan"), None)

            test_o.assertEquals(
                len(interface_o.tagged_vlans.all()),
                len(interface_value.get("tagged_vlans")),
            )
            for vlan_o in interface_o.tagged_vlans.all():
                test_o.assertIn(vlan_o.vid, interface_value.get("tagged_vlans"))

            test_o.assertEquals(
                len(interface_o.ip_addresses.all()),
                len(interface_value.get("ip_addresses")),
            )
            for ipaddress_o in interface_o.ip_addresses.all():
                test_o.assertIn(
                    str(ipaddress_o.address), interface_value.get("ip_addresses")
                )


def test_ipaddresses(test_o, expected_results):
    """Test IPAddress given an expected_results dict."""
    # Test total IPAddress objects
    ipaddress_qs = IPAddress.objects.all()
    test_o.assertEquals(len(ipaddress_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        ipaddress_o = IPAddress.objects.get(address=expected_result.get("address"))
        if ipaddress_o.vrf:
            test_o.assertEquals(ipaddress_o.vrf.name, expected_result.get("vrf"))
        else:
            test_o.assertIs(expected_result.get("vrf"), None)


def test_prefixes(test_o, expected_results):
    """Test Prefix given an expected_results dict."""
    # Test total Prefix objects
    prefix_qs = Prefix.objects.all()
    test_o.assertEquals(len(prefix_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        prefix_o = Prefix.objects.get(prefix=expected_result.get("prefix"))
        if prefix_o.vrf:
            test_o.assertEquals(prefix_o.vrf.name, expected_result.get("vrf"))
        else:
            test_o.assertIs(expected_result.get("vrf"), None)

        if prefix_o.site:
            test_o.assertEquals(prefix_o.site.name, expected_result.get("site"))
        else:
            test_o.assertIs(expected_result.get("site"), None)


def test_vlans(test_o, expected_results):
    """Test VLAN given an expected_results dict."""
    # Test total VLAN objects
    vlan_qs = VLAN.objects.all()
    test_o.assertEquals(len(vlan_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        vlan_o = VLAN.objects.get(vid=int(expected_result.get("vid")))
        test_o.assertEquals(vlan_o.name, expected_result.get("name"))


def test_vrfs(test_o, expected_results):
    """Test VRF given an expected_results dict."""
    # Test total VRF objects
    vrf_qs = VRF.objects.all()
    test_o.assertEquals(len(vrf_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        vrf_o = VRF.objects.get(name=expected_result.get("name"))
        test_o.assertEquals(vrf_o.rd, expected_result.get("rd"))


def test_macaddresses(test_o, expected_results):
    """Test VRF given an expected_results dict."""
    # Test total MacAddressTableEntry objects
    macaddress_qs = MacAddressTableEntry.objects.all()
    test_o.assertEquals(len(macaddress_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        macaddress_o = MacAddressTableEntry.objects.get(
            interface__device__name=expected_result.get("device"),
            interface__label=expected_result.get("interface"),
            mac_address=expected_result.get("mac_address"),
            vvid=expected_result.get("vlan"),
        )
        test_o.assertEquals(macaddress_o.vendor, expected_result.get("vendor"))


def test_arps(test_o, expected_results):
    """Test ArpTableEntry given an expected_results dict."""
    # Test total ArpTableEntry objects
    arp_qs = ArpTableEntry.objects.all()
    test_o.assertEquals(len(arp_qs), len(expected_results))

    # Test each item
    for expected_result in expected_results:
        arp_o = ArpTableEntry.objects.get(
            interface__device__name=expected_result.get("device"),
            interface__label=expected_result.get("interface"),
            mac_address=expected_result.get("mac_address"),
            ip_address=expected_result.get("ip_address") + "/32",
        )
        test_o.assertEquals(arp_o.vendor, expected_result.get("vendor"))


def load_scenario(lab_path):
    """Load DiscoveryLog files and return the list of expected result files."""
    expected_result_files = []

    # Purge all data
    print("Deleting old data... ", end="")
    Cable.objects.all().delete()
    CableTermination.objects.all().delete()
    CablePath.objects.all().delete()  # pylint: disable=no-member
    Device.objects.all().delete()
    DeviceRole.objects.all().delete()
    DeviceType.objects.all().delete()
    Manufacturer.objects.all().delete()
    Prefix.objects.all().delete()
    IPAddress.objects.all().delete()
    Interface.objects.all().delete()
    ArpTableEntry.objects.all().delete()
    MacAddressTableEntry.objects.all().delete()
    RouteTableEntry.objects.all().delete()
    VRF.objects.all().delete()
    VLAN.objects.all().delete()
    DiscoveryLog.objects.all().delete()
    Discoverable.objects.all().delete()
    Credential.objects.all().delete()
    Site.objects.all().delete()
    print("done")

    # Load scenario
    print(f"Loading scenario {lab_path}... ", end="")
    (
        credential_o,
        created,  # pylint: disable=unused-variable
    ) = credential_api.get_or_create(name="test-credential")
    site_o = Site.objects.create(name="test-site")

    for dirpath, dirnames, filenames in walk(  # pylint: disable=unused-variable
        lab_path
    ):
        for filename in filenames:
            if filename.endswith(".yml"):
                expected_result_files.append(f"{dirpath}/{filename}")
            if filename.endswith(".json"):
                filepath_medata = f"{dirpath}/{filename}"
                filepath_raw_output = filepath_medata.replace(".json", ".raw")
                address = filename.split("-")[0]
                with open(filepath_medata, "r", encoding="utf-8") as log_fh:
                    # Load metadata
                    discoverablelog_json = json.load(log_fh)
                with open(filepath_raw_output, "r", encoding="utf-8") as raw_fh:
                    # Load raw output
                    raw_output = raw_fh.read()

                # Create discoverable
                discoverable_o, created = discoverable_api.get_or_create(
                    address=address,
                    discoverable=False,
                    credential_id=credential_o.pk,
                    mode=discoverablelog_json.get("discoverable__mode"),
                    site_id=site_o.pk,
                )

                # Clear metadata and add raw output
                discoverablelog_json["raw_output"] = raw_output
                del discoverablelog_json["discoverable__mode"]

                # Create discoverylog
                discoverylog_api.create(
                    discoverable_id=discoverable_o.pk, **discoverablelog_json
                )
    print("done")

    # Ingest
    print("Ingesting... ", end="")
    logs_qs = DiscoveryLog.objects.filter(parsed=True, ingested=False).order_by("order")
    for log_o in logs_qs:
        log_ingest(log_o)
        log_o.ingested = True
        log_o.save()
    print("done")

    return expected_result_files


class QuestionModelTests(TestCase):
    """Automated test for Netdoc."""

    def test_scenario(self):
        """Run tests on each scenario."""
        for dirpath, dirnames, filenames in walk(  # pylint: disable=unused-variable
            TEST_DIRECTORY
        ):
            if re.match(r".*/lab\d+$", dirpath):
                # Found a test scenario (lab)
                # Load output files (DiscoveryLog)
                expected_result_files = load_scenario(dirpath)

                # Run test
                for expected_result_file in expected_result_files:
                    test_case = (
                        "test_" + expected_result_file.split("/")[-1].split(".")[0]
                    )
                    print(
                        f"Running {test_case} with file {expected_result_file}... ",
                        end="",
                    )
                    with open(expected_result_file, encoding="utf-8") as metadata_fh:
                        expected_results = yaml.load(metadata_fh, Loader=SafeLoader)
                    test_function = globals()[test_case]
                    test_function(self, expected_results)
                    print("done")



"""Ingestor for netmiko_cisco_nxos_show_vrf."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import vrf


def ingest(log):
    """Processing parsed output.

    VRF - Interface association is ingested in the "show ip interface" output.
    """
    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_vrf # pylint: disable=line-too-long
        vrf_name = item.get("name")

        # Get or create VRF
        vrf_o = None
        if vrf_name and vrf_name != "default":
            vrf_o = vrf.get(name=vrf_name)
            if not vrf_o:
                data = {
                    "name": vrf_name,
                }
                vrf_o = vrf.create(**data)

    # Update the log
    log.ingested = True
    log.save()



"""Main class."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"
__version__ = "0.9.54"

import os
import pkgutil
import shutil

from django.conf import settings

from extras.plugins import PluginConfig


PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("netdoc", {})


class NetdocConfig(PluginConfig):
    """Configuration class."""

    name = "netdoc"
    verbose_name = "NetDoc"
    description = "Automatic Network Documentation plugin for NetBox"
    version = __version__
    author = "Andrea Dainese"
    author_email = "andrea@adainese.it"
    base_url = "netdoc"
    required_settings = ["NTC_TEMPLATES_DIR"]
    default_settings = {
        "NTC_TEMPLATES_DIR": "/opt/ntc-templates/ntc_templates/templates",
        "NORNIR_LOG": f"{settings.BASE_DIR}/nornir.log",
        "NORNIR_TIMEOUT": 120,
        "NORNIR_SKIP_LIST": [],
    }


config = NetdocConfig  # pylint: disable=invalid-name

# Setting NTC_TEMPLATES_DIR
os.environ.setdefault("NET_TEXTFSM", PLUGIN_SETTINGS.get("NTC_TEMPLATES_DIR"))

# Copy scripts
package = pkgutil.get_loader("netdoc")
MODULE_PATH = os.path.dirname(package.path)
SCRIPTS_PATH = os.path.join(MODULE_PATH, "scripts")
for filename in os.listdir(SCRIPTS_PATH):
    src_file = os.path.join(SCRIPTS_PATH, filename)
    dst_file = os.path.join(settings.SCRIPTS_ROOT, filename)
    if (
        filename.startswith("__init__")
        or not filename.endswith(".py")
        or not os.path.isfile(src_file)
    ):
        # Not a script file
        continue
    # Copy file in Netbox root scripts path
    shutil.copy(src_file, dst_file)

# Copy reports
REPORTS_PATH = os.path.join(MODULE_PATH, "reports")
for filename in os.listdir(REPORTS_PATH):
    src_file = os.path.join(REPORTS_PATH, filename)
    dst_file = os.path.join(settings.REPORTS_ROOT, filename)
    if (
        filename.startswith("__init__")
        or not filename.endswith(".py")
        or not os.path.isfile(src_file)
    ):
        # Not a report file
        continue
    # Copy file in Netbox root reports path
    shutil.copy(src_file, dst_file)



"""URLs."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from django.urls import path

from netbox.views.generic import ObjectChangeLogView
from netdoc import models, views


urlpatterns = (
    #
    # ARPEntry urls
    #
    path("arptable/", views.ArpTableListView.as_view(), name="arptableentry_list"),
    path("arptable/<int:pk>/", views.ArpTableView.as_view(), name="arptableentry"),
    #
    # Credential urls
    #
    path("credential/", views.CredentialListView.as_view(), name="credential_list"),
    path("credential/add/", views.CredentialEditView.as_view(), name="credential_add"),
    path(
        "credential/import/",
        views.CredentialBulkImportView.as_view(),
        name="credential_import",
    ),
    path(
        "credential/edit/",
        views.CredentialBulkEditView.as_view(),
        name="credential_bulk_edit",
    ),
    path(
        "credential/delete/",
        views.CredentialBulkDeleteView.as_view(),
        name="credential_bulk_delete",
    ),
    path("credential/<int:pk>/", views.CredentialView.as_view(), name="credential"),
    path(
        "credential/<int:pk>/edit/",
        views.CredentialEditView.as_view(),
        name="credential_edit",
    ),
    path(
        "credential/<int:pk>/delete/",
        views.CredentialDeleteView.as_view(),
        name="credential_delete",
    ),
    path(
        "credential/<int:pk>/changelog/",
        ObjectChangeLogView.as_view(),
        name="credential_changelog",
        kwargs={"model": models.Credential},
    ),
    #
    # Diagram urls
    #
    path("diagram/", views.DiagramListView.as_view(), name="diagram_list"),
    path("diagram/add/", views.DiagramEditView.as_view(), name="diagram_add"),
    path("diagram/<int:pk>/", views.DiagramView.as_view(), name="diagram"),
    path(
        "diagram/<int:pk>/edit/",
        views.DiagramEditView.as_view(),
        name="diagram_edit",
    ),
    path(
        "diagram/<int:pk>/export/",
        views.DiagramExportView.as_view(),
        name="diagram_export",
    ),
    path(
        "diagram/<int:pk>/delete/",
        views.DiagramDeleteView.as_view(),
        name="diagram_delete",
    ),
    path(
        "diagram/<int:pk>/changelog/",
        ObjectChangeLogView.as_view(),
        name="diagram_changelog",
        kwargs={"model": models.Diagram},
    ),
    #
    # Discoverable urls
    #
    path(
        "discoverable/", views.DiscoverableListView.as_view(), name="discoverable_list"
    ),
    path(
        "discoverable/add/",
        views.DiscoverableEditView.as_view(),
        name="discoverable_add",
    ),
    path(
        "discoverable/import/",
        views.DiscoverableBulkImportView.as_view(),
        name="discoverable_import",
    ),
    path(
        "discoverable/edit/",
        views.DiscoverableBulkEditView.as_view(),
        name="discoverable_bulk_edit",
    ),
    path(
        "discoverable/delete/",
        views.DiscoverableBulkDeleteView.as_view(),
        name="discoverable_bulk_delete",
    ),
    path(
        "discoverable/discover/",
        views.DiscoverableBulkDiscoverView.as_view(),
        name="discoverable_bulk_discover",
    ),
    path(
        "discoverable/<int:pk>/", views.DiscoverableView.as_view(), name="discoverable"
    ),
    path(
        "discoverable/<int:pk>/edit/",
        views.DiscoverableEditView.as_view(),
        name="discoverable_edit",
    ),
    path(
        "discoverable/<int:pk>/delete/",
        views.DiscoverableDeleteView.as_view(),
        name="discoverable_delete",
    ),
    path(
        "discoverable/<int:pk>/discover/",
        views.DiscoverableDiscoverView.as_view(),
        name="discoverable_discover",
    ),
    path(
        "discoverable/<int:pk>/changelog/",
        ObjectChangeLogView.as_view(),
        name="discoverable_changelog",
        kwargs={"model": models.Discoverable},
    ),
    #
    # DiscoveryLog urls
    #
    path(
        "discoverylog/", views.DiscoveryLogListView.as_view(), name="discoverylog_list"
    ),
    path(
        "discoverylog/delete/",
        views.DiscoveryLogBulkDeleteView.as_view(),
        name="discoverylog_bulk_delete",
    ),
    path(
        "discoverylog/<int:pk>/", views.DiscoveryLogView.as_view(), name="discoverylog"
    ),
    path(
        "discoverylog/<int:pk>/export/",
        views.DiscoveryLogExportView.as_view(),
        name="discoverylog_export",
    ),
    path(
        "discoverylog/<int:pk>/delete/",
        views.DiscoveryLogDeleteView.as_view(),
        name="discoverylog_delete",
    ),
    #
    # MacAddressTableEntry urls
    #
    path(
        "macaddresstable/",
        views.MacAddressTableListView.as_view(),
        name="macaddresstableentry_list",
    ),
    path(
        "macaddresstable/<int:pk>/",
        views.MacAddressTableView.as_view(),
        name="macaddresstableentry",
    ),
    #
    # RoutingTableEntry urls
    #
    path(
        "routingtable/",
        views.RouteTableEntryListView.as_view(),
        name="routingtable_list",
    ),
)



"""Schema validation for VRF."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from ipam.models import VRF

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate VRF data."""
    return {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
            },
            "rd": {
                "type": "string",
            },
            "mandatory_rd": {
                "type": "boolean",
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new VRF objects."""
    schema = get_schema()
    schema["required"] = [
        "name",
    ]
    return schema


def create(mandatory_rd=True, **kwargs):
    """Create an VRF."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    if not mandatory_rd and kwargs.get("rd"):
        # RD must be unique on Netbox, if a VRF exist with the same RD,
        # RD is removed to avoid database key error
        vrf_list = get_list(rd=kwargs.get("rd"))
        if vrf_list and kwargs.get("name") != vrf_list[0].name:
            # Duplicated RD with different VRF name
            del kwargs["rd"]
    obj = utils.object_create(VRF, **kwargs)
    return obj


def get(name):
    """Return an VRF."""
    obj = utils.object_get_or_none(VRF, name=name)
    return obj


def get_list(**kwargs):
    """Get a list of VRF objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(VRF, **kwargs)
    return result


def update(obj, mandatory_rd=True, **kwargs):
    """Update an VRF."""
    update_if_not_set = ["rd"]

    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    if not mandatory_rd and kwargs.get("rd"):
        # RD must be unique on Netbox, if a VRF exist with the same RD,
        # RD is removed to avoid database key error
        vrf_list = get_list(rd=kwargs.get("rd"))
        if vrf_list and kwargs.get("name") != vrf_list[0].name:
            # Duplicated RD with different VRF name
            del kwargs["rd"]
    kwargs_if_not_set = utils.filter_keys(kwargs, update_if_not_set)
    obj = utils.object_update(obj, **kwargs_if_not_set, force=False)
    return obj



"""Ingestor for netmiko_cisco_xr_show_vrf_all_detail."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import vrf


def ingest(log):
    """Processing parsed output.

    VRF - Interface association is ingested in the "show ip interface" output.
    """
    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_vrf # pylint: disable=line-too-long
        vrf_name = item.get("vrf")
        vrf_rd = item.get("rd") if item.get("rd") != "not set" else None

        data = {
            "name": vrf_name,
            "rd": vrf_rd,
        }
        vrf_o = vrf.get(name=vrf_name)
        if vrf_o:
            vrf_o = vrf.update(vrf_o, mandatory_rd=False, **data)
        else:
            vrf_o = vrf.create(mandatory_rd=False, **data)

    # Update the log
    log.ingested = True
    log.save()



"""Ingestor for netmiko_cisco_ios_show_etherchannel_summary."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_etherchannel_summary # pylint: disable=line-too-long
        bundle_name = item.get("po_name")
        bundle_label = utils.normalize_interface_label(bundle_name)
        attached_interface_names = item.get("interfaces")

        # Get or create bundle Interface
        bundle_o = interface.get(device_id=device_o.id, label=bundle_label)
        if not bundle_o:
            bundle_data = {
                "name": bundle_name,
                "device_id": device_o.id,
            }
            bundle_o = interface.create(**bundle_data)
        # Set type on bundle Interface
        bundle_o = interface.update(bundle_o, type="lag")

        for attached_interface_name in attached_interface_names:
            # Get or create attached Interface
            attached_interface_label = utils.normalize_interface_label(
                attached_interface_name
            )
            attached_interface_o = interface.get(
                device_id=device_o.id, label=attached_interface_label
            )
            if not attached_interface_name:
                attached_interface_data = {
                    "name": attached_interface_name,
                    "device_id": device_o.id,
                }
                attached_interface_o = interface.create(**attached_interface_data)
            # Set LAG on attached Interface
            attached_interface_o = interface.update(
                attached_interface_o, lag_id=bundle_o.id
            )

    # Update the log
    log.ingested = True
    log.save()



"""Ingestor for netmiko_cisco_nxos_show_interface."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_interface # pylint: disable=line-too-long
        interface_name = item.get("interface")
        label = utils.normalize_interface_label(interface_name)
        description = item.get("description")
        duplex = utils.normalize_interface_duplex(item.get("duplex"))
        speed = utils.normalize_interface_speed(item.get("bandwidth"))
        mac_address = (
            utils.normalize_mac_address(item.get("address"))
            if not utils.incomplete_mac(item.get("address"))
            else None
        )
        int_type = utils.normalize_interface_type(item.get("interface"))
        enabled = utils.normalize_interface_status(item.get("link_status"))
        mtu = utils.normalize_interface_mtu(item.get("mtu"))
        parent_name = utils.parent_interface(label)

        if parent_name:
            # Parent Interface is set
            parent_label = utils.normalize_interface_label(parent_name)
            parent_o = interface.get(device_id=device_o.id, label=parent_label)
            if not parent_o:
                parent_data = {
                    "name": parent_name,
                    "device_id": device_o.id,
                }
                parent_o = interface.create(**parent_data)

        data = {
            "name": interface_name,
            "description": description,
            "duplex": duplex,
            "speed": speed,
            "mac_address": mac_address,
            "type": int_type,
            "device_id": device_o.id,
            "enabled": enabled,
            "mtu": mtu,
            "parent_id": parent_o.id if parent_name else None,
        }
        interface_o = interface.get(device_id=device_o.id, label=label)
        if not interface_o:
            interface.create(**data)
        else:
            interface.update(interface_o, **data)

    # Update the log
    log.ingested = True
    log.save()



"""Schema validation for Discoverable."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from django.utils import timezone

from dcim.models import Device, Site

from netdoc.models import Discoverable, Credential, DiscoveryModeChoices
from netdoc import utils


def get_schema():
    """Return the JSON schema to validate Discoverable data."""
    return {
        "type": "object",
        "properties": {
            "address": {
                "type": "string",
                "format": "ipv4",
            },
            "device_id": {
                "type": "integer",
                "enum": list(Device.objects.all().values_list("id", flat=True)),
            },
            "credential_id": {
                "type": "integer",
                "enum": list(Credential.objects.all().values_list("id", flat=True)),
            },
            "discoverable": {
                "type": "boolean",
            },
            "last_discovered_at": {
                "type": "string",
                "format": "date-time",
            },
            "mode": {
                "type": "string",
                "enum": [key for key, value in DiscoveryModeChoices()],
            },
            "site_id": {
                "type": "integer",
                "enum": list(Site.objects.all().values_list("id", flat=True)),
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Discoverable objects."""
    schema = get_schema()
    schema["required"] = [
        "address",
        "credential_id",
        "mode",
        "site_id",
    ]
    return schema


def create(**kwargs):
    """Create a Discoverable."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(Discoverable, **kwargs)
    return obj


def get(address, discovered=False):
    """Return a Discoverable given IP address.

    Update last_discovered_at if required.
    """
    obj = utils.object_get_or_none(Discoverable, address=address)
    if obj and discovered:
        obj = utils.object_update(obj, last_discovered_at=timezone.now())
    return obj


def get_list(**kwargs):
    """Get a list of Discoverable objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Discoverable, **kwargs)
    return result


def get_or_create(address, **kwargs):
    """Get or create a Discoverable."""
    created = False
    data = {
        **kwargs,
        "address": address,
    }
    data = utils.delete_empty_keys(data)
    validate(data, get_schema_create(), format_checker=FormatChecker())

    obj = utils.object_get_or_none(Discoverable, address=address)
    if not obj:
        obj = utils.object_create(Discoverable, address=address, **kwargs)
        created = True

    return obj, created


def update(obj, **kwargs):
    """Update a Discoverable."""
    update_if_not_set = ["device_id"]
    update_always = ["last_discovered_at"]

    validate(kwargs, get_schema(), format_checker=FormatChecker())
    kwargs_if_not_set = utils.filter_keys(kwargs, update_if_not_set)
    kwargs_always = utils.filter_keys(kwargs, update_always)
    obj = utils.object_update(obj, **kwargs_always, force=True)

    # Before updating device_id we need to check for integrity
    if kwargs_if_not_set.get("device_id"):
        # Check if the device is already associated
        obj_qs = Discoverable.objects.filter(
            device_id=kwargs_if_not_set.get("device_id")
        )
        if obj_qs:
            # A Discoverable with the same address or the same Device ID already exist
            return obj_qs[0]
    obj = utils.object_update(obj, **kwargs_if_not_set, force=False)
    return obj



"""Serializers, called by API Views for add/ediit actions."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from rest_framework import serializers

from netbox.api.serializers import NetBoxModelSerializer

from netdoc.models import (
    Credential,
    Discoverable,
    DiscoveryLog,
    RouteTableEntry,
    ArpTableEntry,
    MacAddressTableEntry,
    Diagram,
)


class ArpTableEntrySerializer(NetBoxModelSerializer):
    """Serializer to validate ArpTableEntry data.

    No need to view-details/edit/delete, so url and View are missing.
    """

    class Meta:
        """Serializer metadata."""

        model = ArpTableEntry
        fields = "__all__"


class CredentialSerializer(NetBoxModelSerializer):
    """Serializer to validate Credential data."""

    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netdoc-api:credential-detail"
    )
    discoverables_count = serializers.IntegerField(read_only=True)

    class Meta:
        """Serializer metadata."""

        model = Credential
        fields = "__all__"


class DiagramSerializer(NetBoxModelSerializer):
    """Serializer to validate Diagram data."""

    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netdoc-api:diagram-detail"
    )
    discoverables_count = serializers.IntegerField(read_only=True)

    class Meta:
        """Serializer metadata."""

        model = Diagram
        fields = "__all__"


class DiscoverableSerializer(NetBoxModelSerializer):
    """Serializer to validate Discoverable data."""

    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netdoc-api:discoverable-detail"
    )

    class Meta:
        """Serializer metadata."""

        model = Discoverable
        fields = "__all__"


class DiscoveryLogSerializer(NetBoxModelSerializer):
    """Serializer to validate DiscoveryLog data."""

    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netdoc-api:discoverylog-detail"
    )

    class Meta:
        """Serializer metadata."""

        model = DiscoveryLog
        fields = "__all__"


class MacAddressTableEntrySerializer(NetBoxModelSerializer):
    """Serializer to validate MacAddressTableEntry data.

    No need to view-details/edit/delete, so url and View are missing.
    """

    class Meta:
        """Serializer metadata."""

        model = MacAddressTableEntry
        fields = "__all__"


class RouteTableEntrySerializer(NetBoxModelSerializer):
    """Serializer to validate RouteTableEntry data.

    No need to view-details/edit/delete, so url and View are missing.
    """

    class Meta:
        """Serializer metadata."""

        model = RouteTableEntry
        fields = "__all__"



"""Schema validation for Manufacturer."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker
from slugify import slugify

from dcim.models import Manufacturer

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate Manufacturer data."""
    return {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
            },
            "description": {
                "type": "string",
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Manufacturer objects."""
    schema = get_schema()
    schema["required"] = [
        "name",
    ]
    return schema


def create(**kwargs):
    """Create a Manufacturer."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    kwargs["slug"] = slugify(kwargs.get("name"))
    obj = utils.object_create(Manufacturer, **kwargs)
    return obj


def get(name):
    """Return a Manufacturer."""
    obj = utils.object_get_or_none(Manufacturer, name=name)
    return obj


def get_list(**kwargs):
    """Get a list of Manufacturer objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Manufacturer, **kwargs)
    return result



"""Ingestor for netmiko_hp_comware_display_interface."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/hp_comware/display_interface # pylint: disable=line-too-long
        interface_name = item.get("intf")
        label = utils.normalize_interface_label(interface_name)
        description = item.get("description")
        duplex = utils.normalize_interface_duplex(item.get("duplex"))
        speed = utils.normalize_interface_speed(item.get("bandwidth"))
        mac_address = (
            utils.normalize_mac_address(item.get("hw_address")[0])
            if item.get("hw_address")
            and not utils.incomplete_mac(item.get("hw_address")[0])
            else None
        )
        mode = utils.normalize_interface_mode(item.get("port_link_type"))
        int_type = utils.normalize_interface_type(item.get("intf"))
        enabled = (
            utils.normalize_interface_status(item.get("line_status"))
            if item.get("line_status")
            else None
        )
        mtu = utils.normalize_interface_mtu(item.get("mtu"))
        parent_name = utils.parent_interface(label)
        native_vlan = int(item.get("vlan_native")) if item.get("vlan_native") else None
        tagged_vlans = utils.normalize_vlan_list(item.get("vlan_permitted"))

        if parent_name:
            # Parent Interface is set
            parent_label = utils.normalize_interface_label(parent_name)
            parent_o = interface.get(device_id=device_o.id, label=parent_label)
            if not parent_o:
                parent_data = {
                    "name": parent_name,
                    "device_id": device_o.id,
                }
                parent_o = interface.create(**parent_data)

        data = {
            "name": interface_name,
            "description": description,
            "duplex": duplex,
            "speed": speed,
            "mac_address": mac_address,
            "type": int_type,
            "device_id": device_o.id,
            "enabled": enabled,
            "mtu": mtu,
            "parent_id": parent_o.id if parent_name else None,
        }

        interface_o = interface.get(device_id=device_o.id, label=label)
        if not interface_o:
            interface_o = interface.create(**data)
        else:
            interface_o = interface.update(interface_o, **data)

        # Update Interface mode
        mode_data = {
            "untagged_vlan": native_vlan,
            "mode": mode,
            "tagged_vlans": tagged_vlans,
        }
        interface.update_mode(interface_o, **mode_data)

    # Update the log
    log.ingested = True
    log.save()



"""Tables, called by Views."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import django_tables2 as tables

from netbox.tables import NetBoxTable, ChoiceFieldColumn
from netbox.tables.columns import ActionsColumn

from netdoc import models


#
# ArpTableEntry tables
#


class ArpTableEntryTable(NetBoxTable):
    """ARP list table used in ArpTableEntryView."""

    device = tables.LinkColumn(
        "dcim:device",
        accessor="interface.device",
        args=[tables.utils.A("interface__device__pk")],
    )
    interface = tables.LinkColumn(
        "dcim:interface", args=[tables.utils.A("interface__pk")]
    )
    mac_address = tables.LinkColumn(
        "plugins:netdoc:arptableentry", args=[tables.utils.A("pk")]
    )
    device_role = tables.Column(accessor="interface.device.device_role")
    actions = []  # Read only table

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.ArpTableEntry
        fields = [
            "pk",
            "id",
            "device",
            "device_role",
            "interface",
            "ip_address",
            "mac_address",
            "vendor",
            "last_updated",
            "created",
        ]
        default_columns = [
            "device",
            "device_role",
            "interface",
            "ip_address",
            "mac_address",
            "vendor",
            "last_updated",
        ]


#
# Credential tables
#


class CredentialTable(NetBoxTable):
    """Credential list table used in CredentialListView."""

    name = tables.Column(linkify=True)
    discoverables_count = tables.Column()

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.Credential
        fields = ["pk", "id", "name", "username", "discoverables_count"]
        default_columns = ["name", "username", "discoverables_count"]


#
# Diagram tables
#


class DiagramTable(NetBoxTable):
    """Diagram list table used in DiagramListView."""

    name = tables.Column(linkify=True)
    discoverables_count = tables.Column()

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.Diagram
        fields = ["pk", "id", "name", "mode", "created", "last_updated"]
        default_columns = ["name", "mode", "created", "last_updated"]


#
# Discoverable tables
#


class DiscoverableTable(NetBoxTable):
    """Credential list table used in DiscoverableListView."""

    address = tables.Column(linkify=True)
    device = tables.Column(linkify=True)
    mode = ChoiceFieldColumn()
    discoverylogs_count = tables.Column()
    discovery_button = """
    <a class="btn btn-sm btn-secondary" href="{% url 'plugins:netdoc:discoverable_discover' pk=record.pk %}" title="Discover">
        <i class="mdi mdi-refresh"></i>
    </a>
    """
    actions = ActionsColumn(extra_buttons=discovery_button)

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.Discoverable
        fields = [
            "pk",
            "id",
            "address",
            "device",
            "site",
            "credential",
            "mode",
            "discoverable",
            "last_discovered_at",
            "discoverylogs_count",
            "last_updated",
            "created",
        ]
        default_columns = [
            "address",
            "device",
            "site",
            "credential",
            "mode",
            "discoverable",
            "last_discovered_at",
            "discoverylogs_count",
        ]


class DiscoverableTableWOLogCount(NetBoxTable):
    """Credential list table used in DiscoverableListView.

    DiscoveryLog count is omitted. Used in CredentialView.
    """

    address = tables.Column(linkify=True)
    mode = ChoiceFieldColumn()
    discovery_button = """
    <a class="btn btn-sm btn-secondary" href="{% url 'plugins:netdoc:discoverable_discover' pk=record.pk %}" title="Discover">
        <i class="mdi mdi-refresh"></i>
    </a>
    """
    actions = ActionsColumn(extra_buttons=discovery_button)

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.Discoverable
        fields = [
            "pk",
            "id",
            "address",
            "device",
            "site",
            "credential",
            "mode",
            "discoverable",
            "last_discovered_at",
            "last_updated",
            "created",
        ]
        default_columns = [
            "address",
            "device",
            "site",
            "credential",
            "mode",
            "discoverable",
            "last_discovered_at",
        ]


#
# Discovery log tables
#


class DiscoveryLogTable(NetBoxTable):
    """Credential list table used in DiscoveryLogListView."""

    device = tables.Column(linkify=True, accessor="discoverable.device")
    discoverable = tables.Column(linkify=True)
    actions = ActionsColumn(actions=("delete",))  # Read only + delete table

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.DiscoveryLog
        fields = [
            "pk",
            "id",
            "created",
            "discoverable",
            "device",
            "command",
            "template",
            "order",
            "configuration",
            "supported",
            "success",
            "parsed",
            "ingested",
            "last_updated",
        ]
        default_columns = [
            "id",
            "created",
            "discoverable",
            "device",
            "command",
            "order",
            "configuration",
            "supported",
            "success",
            "parsed",
            "ingested",
        ]


#
# MacAddressTableEntry tables
#


class MacAddressTableEntryTable(NetBoxTable):
    """MAC address list table used in MacAddressTableListView."""

    device = tables.LinkColumn(
        "dcim:device",
        accessor="interface.device",
        args=[tables.utils.A("interface__device__pk")],
    )
    interface = tables.LinkColumn(
        "dcim:interface", args=[tables.utils.A("interface__pk")]
    )
    mac_address = tables.LinkColumn(
        "plugins:netdoc:macaddresstableentry", args=[tables.utils.A("pk")]
    )
    device_role = tables.Column(accessor="interface.device.device_role")
    actions = []  # Read only table

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.MacAddressTableEntry
        fields = [
            "pk",
            "id",
            "device",
            "device_role",
            "interface",
            "vvid",
            "mac_address",
            "vendor",
            "last_updated",
            "created",
        ]
        default_columns = [
            "device",
            "device_role",
            "interface",
            "vvid",
            "mac_address",
            "vendor",
            "last_updated",
        ]


#
# RouteTableEntry tables
#


class RouteTableEntryTable(NetBoxTable):
    """Route list table used in RouteTableEntryListView."""

    device = tables.LinkColumn(
        "dcim:device",
        args=[tables.utils.A("device__pk")],
    )
    nexthop_if = tables.LinkColumn(
        "dcim:interface", args=[tables.utils.A("nexthop_if__pk")]
    )
    device_role = tables.Column(accessor="device.device_role")
    actions = []  # Read only table

    class Meta(NetBoxTable.Meta):
        """Table metadata."""

        model = models.RouteTableEntry
        fields = [
            "pk",
            "id",
            "device",
            "device_role",
            "destination",
            "protocol",
            "distance",
            "metric",
            "nexthop_ip",
            "nexthop_if",
            "vrf",
            "last_updated",
            "created",
        ]
        default_columns = [
            "device",
            "destination",
            "protocol",
            "distance",
            "metric",
            "nexthop_ip",
            "nexthop_if",
            "vrf",
            "last_updated",
        ]



"""Discovery task for HP Comware devices via Netmiko."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import json
from nornir_utils.plugins.functions import print_result
from nornir.core.filter import F

from netdoc import utils
from netdoc.schemas import discoverable, discoverylog


def discovery(nrni):
    """Discovery Cisco IOS devices."""
    platform = "hp_comware"
    filtered_devices = nrni.filter(platform=platform)

    def multiple_tasks(task):
        """Define commands (in order) for the playbook."""
        utils.append_nornir_netmiko_task(
            task,
            "display current-configuration | include sysname",
            template="HOSTNAME",
            order=0,
        )
        utils.append_nornir_netmiko_task(
            task,
            [
                "display current-configuration",
                # Depending on version, "verbose" may be unsupportted
                "display lldp neighbor-information verbose",
                "display lldp neighbor-information list",
                # Depending on version, "brief" may be unsupportted
                "display vlan brief",
                "display vlan all",
                "display ip vpn-instance",
            ],
            order=10,
        )
        utils.append_nornir_netmiko_task(
            task,
            [
                "display interface",
                "display ip interface",
            ],
            order=50,
        )
        utils.append_nornir_netmiko_task(
            task,
            [
                "display mac-address",
                "display link-aggregation verbose",
                "display_device_manuinfo",
            ],
        )
        utils.append_nornir_netmiko_task(
            task,
            [
                "display version",
                "display logbuffer level 6",
                "display stp",
                "display port trunk",
                "display vrrp",
                "display ospf peer",
                "display bgp peer",
            ],
            supported=False,
        )

    # Run the playbook
    aggregated_results = filtered_devices.run(task=multiple_tasks)

    # Print the result
    print_result(aggregated_results)

    # Save outputs and define additional commands
    for key, multi_result in aggregated_results.items():
        vrfs = ["default"]  # Adding default VRF
        current_nr = nrni.filter(F(name=key))

        # MultiResult is an array of Result
        for result in multi_result:
            if result.name == "multiple_tasks":
                # Skip MultipleTask
                continue

            address = result.host.dict().get("hostname")
            discoverable_o = discoverable.get(address, discovered=True)
            details = json.loads(result.name)
            discoverylog.create(
                command=details.get("command"),
                discoverable_id=discoverable_o.id,
                raw_output=result.result,
                template=details.get("template"),
                order=details.get("order"),
                details=details,
            )

            # Save VRF list for later
            if details.get("command") == "display ip vpn-instance":
                parsed_vrfs, parsed = utils.parse_netmiko_output(
                    result.result, details.get("command"), platform
                )
                if parsed:
                    for vrf in parsed_vrfs:
                        vrfs.append(vrf.get("name"))

        # Additional commands out of the multi result loop
        def additional_tasks(task):
            """Define additional commands (in order) for the playbook."""
            # Per VRF commands
            for vrf in vrfs:  # pylint: disable=cell-var-from-loop
                if vrf == "default":
                    # Default VRF has no name
                    utils.append_nornir_netmiko_task(
                        task,
                        [
                            "display arp",
                            "display ip routing-table",
                        ],
                    )
                else:
                    # with non default VRF commands and templates differ
                    utils.append_nornir_netmiko_task(
                        task,
                        commands=f"display arp vpn-instance {vrf}",
                        template="display arp",
                    )
                    utils.append_nornir_netmiko_task(
                        task,
                        commands=f"display ip vpn-instance instance-name {vrf}",
                        template="display ip vpn-instance instance-name",
                    )
                    utils.append_nornir_netmiko_task(
                        task,
                        commands=f"display ip routing-table vpn-instance {vrf}",
                        template="display ip routing-table",
                    )

        # Run the additional playbook
        additional_aggregated_results = current_nr.run(task=additional_tasks)

        # Print the result
        print_result(additional_aggregated_results)

        for key, additional_multi_result in additional_aggregated_results.items():
            # MultiResult is an array of Result
            for result in additional_multi_result:
                if result.name == "additional_tasks":
                    # Skip MultipleTask
                    continue

                details = json.loads(result.name)
                if " vpn-instance " in details.get("command"):
                    # Save the VRF in details
                    details["vrf"] = (
                        details.get("command").split(" vpn-instance ").pop()
                    )
                elif " instance-name " in details.get("command"):
                    # Save the VRF in details
                    details["vrf"] = (
                        details.get("command").split(" instance-name ").pop()
                    )
                discoverylog.create(
                    command=details.get("command"),
                    discoverable_id=discoverable_o.id,
                    raw_output=result.result,
                    template=details.get("template"),
                    order=details.get("order"),
                    details=details,
                )



"""Models (ORM).

Define ORM models for NetDoc objects:
* A Discoverable is a device willed to be discovered using a specific mode.
* Credential is associated to one or more Discoverables.
* A DiscoveryLog is the output for a specific discovery command executed in a Discoverable.
"""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import re
from OuiLookup import OuiLookup

from django.db import models
from django.urls import reverse

from netbox.models import NetBoxModel
from ipam.fields import IPAddressField
from dcim.fields import MACAddressField
from utilities.choices import ChoiceSet

from netdoc.utils import parse_netmiko_output, CONFIG_COMMANDS, FAILURE_OUTPUT


class DeviceImageChoices(ChoiceSet):
    """Image used in diagrams associated to device roles."""

    CHOICES = [
        ("access-switch", "Access Switch"),
        ("backup", "Backup Device"),
        ("circuit", "Circuit"),
        ("core-switch", "Core Switch"),
        ("distribution-switch", "Distribution Switch"),
        ("firewall", "Firewall"),
        ("internal-switch", "Internal Switch"),
        ("isp-cpe-material", "ISP CPE Material"),
        ("non-racked-devices", "Non Racked Device"),
        ("power-feed", "Power Feed"),
        ("power-panel", "Power Panel"),
        ("power-units", "Power Units"),
        ("provider-networks", "Provider Networks"),
        ("unknown", "Unknown"),
        ("router", "Router"),
        ("server", "Server"),
        ("storage", "Storage"),
        ("wan-network", "WAN Network"),
        ("wireless-ap", "Wireless AP"),
    ]


class DiagramModeChoices(ChoiceSet):
    """Diagram mode."""

    CHOICES = [
        ("l2", "L2"),
        ("l3", "L3"),
        # ("stp", "STP"),
    ]


class DiscoveryModeChoices(ChoiceSet):
    """Discovey mode."""

    CHOICES = [
        ("netmiko_cisco_ios", "Netmiko Cisco IOS XE"),
        ("netmiko_cisco_nxos", "Netmiko Cisco NX-OS"),
        ("netmiko_cisco_xr", "Netmiko Cisco XR"),
        ("netmiko_hp_comware", "Netmiko HPE Comware"),
        ("netmiko_linux", "Netmiko Linux"),
    ]


class RouteTypeChoices(ChoiceSet):
    """Route type."""

    CHOICES = [
        ("u", "Unknown"),
        ("b", "BGP"),
        ("c", "Connected"),
        ("s", "Static"),
        ("r", "RIP"),
        ("e", "EIGRP"),
        ("ex", "EIGRP external"),
        ("oia", "OSPF inter area"),
        ("on1", "OSPF NSSA external type 1"),
        ("on2", "OSPF NSSA external type 2"),
        ("oe1", "OSPF external type 1"),
        ("oe2", "OSPF external type 2"),
        ("i", "IS-IS"),
        ("is", "IS-IS summary"),
        ("i1", "IS-IS level-1"),
        ("i2", "IS-IS level-2"),
    ]


#
# ARPEntry model
#


class ArpTableEntry(NetBoxModel):
    """
    Model for ArpTableEntry.

    Each ARP seen on each network interface is
    counted. One IP Address can be associated to one MAC Address. One MAC
    Address can be associated to multiple IP Addresses.
    """

    interface = models.ForeignKey(
        to="dcim.Interface",
        on_delete=models.CASCADE,
        related_name="+",
        editable=False,
    )
    ip_address = IPAddressField(help_text="IPv4 address", editable=False)
    mac_address = MACAddressField(help_text="MAC Address", editable=False)
    vendor = models.CharField(
        max_length=255, blank=True, null=True, help_text="Vendor", editable=False
    )  #: Vendor (from OUI)

    class Meta:
        """Database metadata."""

        ordering = ["ip_address"]
        unique_together = ["interface", "ip_address", "mac_address"]
        verbose_name = "ARP table entry"
        verbose_name_plural = "ARP table entries"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return f"{self.ip_address} has {self.mac_address} at {self.interface}"

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:arptableentry", args=[self.pk])

    def save(self, *args, **kwargs):
        """Set vendor field when saving."""
        self.vendor = list(
            OuiLookup().query(str(self.mac_address)).pop().values()
        ).pop()
        super().save(*args, **kwargs)


#
# Credential model
#


class Credential(NetBoxModel):
    """Model for Credential."""

    name = models.CharField(max_length=100)
    enable_password = models.CharField(
        max_length=100,
        blank=True,
    )
    password = models.CharField(
        max_length=100,
        blank=True,
    )
    username = models.CharField(
        max_length=100,
        blank=True,
    )

    class Meta:
        """Database metadata."""

        ordering = ["name"]
        unique_together = ["name"]
        verbose_name = "Credential"
        verbose_name_plural = "Credentials"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return str(self.name)

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:credential", args=[self.pk])


#
# Diagram model
#


class Diagram(NetBoxModel):
    """Model for Diagram."""

    details = models.JSONField(
        default=dict,
    )  #: Vis.Js Details in JSON format
    device_roles = models.ManyToManyField(
        to="dcim.DeviceRole",
        related_name="+",
        blank=True,
    )
    mode = models.CharField(
        max_length=30,
        choices=DiagramModeChoices,
    )
    name = models.CharField(max_length=100)
    sites = models.ManyToManyField(
        to="dcim.Site",
        related_name="+",
        blank=True,
    )
    vrfs = models.ManyToManyField(
        to="ipam.VRF",
        related_name="+",
        blank=True,
    )
    include_global_vrf = models.BooleanField(
        default=False
    )  #: Always include Global VRF (None)

    class Meta:
        """Database metadata."""

        ordering = ["name"]
        unique_together = [
            "name",
        ]
        verbose_name = "Diagram"
        verbose_name_plural = "Diagrams"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return str(self.name)

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:diagram", args=[self.pk])


#
# Discoverable model
#


class Discoverable(NetBoxModel):
    """Model for Discoverable."""

    address = models.GenericIPAddressField()
    device = models.OneToOneField(
        to="dcim.Device",
        on_delete=models.SET_NULL,
        related_name="+",
        blank=True,
        null=True,
    )
    credential = models.ForeignKey(
        to=Credential,
        on_delete=models.PROTECT,
        related_name="discoverables",
    )
    mode = models.CharField(
        max_length=30,
        choices=DiscoveryModeChoices,
    )
    discoverable = models.BooleanField(
        default=False
    )  #: New created devices have discoverable=False by default (e.g. if created from CDP/LLDP)
    last_discovered_at = models.DateTimeField(blank=True, null=True, editable=False)
    site = models.ForeignKey(
        to="dcim.Site",
        on_delete=models.CASCADE,
        related_name="+",
    )

    class Meta:
        """Database metadata."""

        ordering = ["mode", "address"]
        unique_together = ["address", "mode"]
        verbose_name = "Device"
        verbose_name_plural = "Devices"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return f"{self.address} via {self.mode}"

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:discoverable", args=[self.pk])


#
# Discovery log model
#


class DiscoveryLog(NetBoxModel):
    """Model for DiscoveryLog."""

    command = models.CharField(
        max_length=255, editable=False
    )  #: Exact CMD used to discover
    configuration = models.BooleanField(
        default=False,
        editable=False,
    )
    discoverable = models.ForeignKey(
        to=Discoverable,
        on_delete=models.CASCADE,
        related_name="discoverylogs",
        editable=False,
    )
    details = models.JSONField(
        default=dict,
        editable=False,
    )  #: Details in JSON format stored within Nornir task name
    order = models.IntegerField(default=128)  #: Order to ingest (lower is sooner)
    parsed_output = models.JSONField(default=list, editable=False)
    raw_output = models.JSONField(
        default=dict,
    )
    template = models.CharField(
        max_length=255, editable=False
    )  #: Template used to ingest parsed_output
    success = models.BooleanField(
        default=False, editable=False
    )  # True if excuting request return OK and raw_output is valid (avoid command not found)
    supported = models.BooleanField(
        default=True, editable=False
    )  #: False if output is unsupported (won't be parsed/ingested)
    parsed = models.BooleanField(
        default=False, editable=False
    )  #: True if parsing raw_output return a valid JSON
    ingested = models.BooleanField(
        default=False, editable=False
    )  #: True if all data are ingested without errors

    class Meta:
        """Database metadata."""

        ordering = ["created"]
        verbose_name = "Log"
        verbose_name_plural = "Logs"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return f"{self.command} at {self.created}"

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:discoverylog", args=[self.pk])

    def parse(self):
        """Parse raw_output."""
        self.configuration = False
        self.success = False
        self.parsed = False
        self.parsed_output = ""
        mode = self.discoverable.mode  # pylint: disable=no-member

        # Check if the command is supported
        if not self.supported:
            return

        # Check if the output is a configuration file
        if self.template != "HOSTNAME":
            for regex in CONFIG_COMMANDS:
                if re.search(regex, self.command):
                    self.configuration = True
                    # Configuraion file -> skip parsing
                    return

        # Check if the output is completed successfully
        for regex in FAILURE_OUTPUT:
            if re.search(regex, self.raw_output):
                # Command failed -> skip parsing
                return
        self.success = True

        # Parse framework (e.g. netmiko) and platform (e.g. cisco_ios)
        framework = mode.split("_").pop(0)
        platform = "_".join(mode.split("_")[1:])

        if framework == "netmiko":
            parsed_output, parsed = parse_netmiko_output(
                self.raw_output, self.command, platform, template=self.template
            )
        else:
            raise ValueError("Framework not detected")

        self.parsed_output = parsed_output
        self.parsed = parsed

    def save(self, *args, **kwargs):
        """Parse raw_output when saving."""
        if "supported" in dict(self.details):
            self.supported = dict(self.details).get("supported")
        if not self.pk:
            # Prase (once) before creating the object
            self.parse()

        super().save(*args, **kwargs)


#
# MacAddressTableEntry model
#


class MacAddressTableEntry(NetBoxModel):
    """Model for MacAddressTableEntry.

    Each MAC Address seen on each network
    interface is counted. One IP Address can be associated to one AC
    Address. One MAC Address can be associated to multiple IP Addresses.
    """

    interface = models.ForeignKey(
        to="dcim.Interface",
        on_delete=models.CASCADE,
        related_name="+",
        editable=False,
    )
    mac_address = MACAddressField(help_text="MAC Address", editable=False)
    vendor = models.CharField(
        max_length=255, blank=True, null=True, help_text="Vendor", editable=False
    )  #: Vendor (from OUI)
    vvid = models.IntegerField(help_text="VLAN ID")  #: VLAN ID (TAG)

    class Meta:
        """Database metadata."""

        ordering = ["mac_address", "vvid"]
        unique_together = ["interface", "mac_address", "vvid"]
        verbose_name = "MAC Address table entry"
        verbose_name_plural = "MAC Address table entries"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        return f"{self.mac_address} is at {self.interface}"

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:macaddresstableentry", args=[self.pk])

    def save(self, *args, **kwargs):
        """Set vendor field when saving."""
        self.vendor = list(
            OuiLookup().query(str(self.mac_address)).pop().values()
        ).pop()
        super().save(*args, **kwargs)


#
# RouteTableEntry model
#


class RouteTableEntry(NetBoxModel):
    """Model for RouteTableEntry.

    Each route has a destination, type (connected, static...),
    nexthop_ip and/or nexthop_if, distance (administrative), metric.
    """

    destination = IPAddressField(help_text="Destination network", editable=False)
    device = models.ForeignKey(
        to="dcim.Device",
        on_delete=models.CASCADE,
        related_name="+",
    )
    distance = models.IntegerField(blank=True, null=True, editable=False)
    metric = models.BigIntegerField(blank=True, null=True, editable=False)
    nexthop_ip = IPAddressField(
        help_text="IPv4 address",
        editable=False,
        blank=True,
        null=True,
    )
    nexthop_if = models.ForeignKey(
        to="dcim.Interface",
        on_delete=models.CASCADE,
        related_name="+",
        editable=False,
        blank=True,
        null=True,
    )
    protocol = models.CharField(
        max_length=30,
        choices=RouteTypeChoices,
        editable=False,
    )
    vrf = models.ForeignKey(
        to="ipam.VRF",
        on_delete=models.CASCADE,
        related_name="+",
        editable=False,
        blank=True,
        null=True,
    )

    class Meta:
        """Database metadata."""

        ordering = ["device", "protocol", "metric"]
        unique_together = [
            "device",
            "destination",
            "distance",
            "metric",
            "protocol",
            "vrf",
        ]
        verbose_name = "Route"
        verbose_name_plural = "Routes"

    def __str__(self):
        """Return a human readable name when the object is printed."""
        if self.nexthop_ip:
            return f"{self.destination} [{self.distance}/{self.metric}] via {self.nexthop_ip}"
        # Assuming nexthop_if
        return (
            f"{self.destination} [{self.distance}/{self.metric}] at {self.nexthop_if}"
        )

    def get_absolute_url(self):
        """Return the absolute url."""
        return reverse("plugins:netdoc:routingtable", args=[self.pk])



"""API URLs, called by Views, used for add/edit actions."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netbox.api.routers import NetBoxRouter
from netdoc.api import views


app_name = "nedoc"  # pylint: disable=invalid-name

router = NetBoxRouter()
router.register("credential", views.CredentialViewSet)
router.register("diagram", views.DiagramViewSet)
router.register("discoverylog", views.DiscoveryLogViewSet)
router.register("discoverable", views.DiscoverableViewSet)

urlpatterns = router.urls



"""Advanced filters."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from django.db.models import Q

from netbox.filtersets import NetBoxModelFilterSet

from netdoc.models import (
    Discoverable,
    Credential,
    DiscoveryLog,
    ArpTableEntry,
    MacAddressTableEntry,
    RouteTableEntry,
)


class ArpTableEntryFilterSet(NetBoxModelFilterSet):
    """FilterSet used in ArpTableListView."""

    class Meta:
        """FilterSet metadata."""

        model = ArpTableEntry
        fields = []

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(
            Q(ip_address__icontains=value)
            | Q(mac_address__icontains=value)
            | Q(interface__name__icontains=value)
            | Q(interface__device__name__icontains=value)
            | Q(vendor__icontains=value)
        )


class CredentialFilterSet(NetBoxModelFilterSet):
    """FilterSet used in Credential."""

    class Meta:
        """FilterSet metadata."""

        model = Credential
        fields = []

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(Q(name__icontains=value) | Q(username__icontains=value))


class DiscoverableFilterSet(NetBoxModelFilterSet):
    """FilterSet used in Discoverable."""

    class Meta:
        """FilterSet metadata."""

        model = Discoverable
        fields = []

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(
            Q(address__icontains=value)
            | Q(device__name__icontains=value)
            | Q(site__name__icontains=value)
            | Q(credential__name__icontains=value)
            | Q(mode__icontains=value)
        )


class DiscoveryLogFilterSet(NetBoxModelFilterSet):
    """FilterSet used in DiscoveryLog."""

    class Meta:
        """FilterSet metadata."""

        model = DiscoveryLog
        fields = ["configuration", "success", "supported", "parsed", "ingested"]

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(
            Q(discoverable__address__icontains=value)
            | Q(discoverable__mode__icontains=value)
            | Q(discoverable__device__name__icontains=value)
            | Q(command__icontains=value)
        )


class MacAddressTableEntryFilterSet(NetBoxModelFilterSet):
    """FilterSet used in MacAddressTableListView."""

    class Meta:
        """FilterSet metadata."""

        model = MacAddressTableEntry
        fields = []

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(
            Q(mac_address__icontains=value)
            | Q(interface__name__icontains=value)
            | Q(interface__device__name__icontains=value)
            | Q(vendor__icontains=value)
        )


class RouteTableEntryFilterSet(NetBoxModelFilterSet):
    """FilterSet used in RouteTableEntryListView."""

    class Meta:
        """FilterSet metadata."""

        model = RouteTableEntry
        fields = []

    def search(self, queryset, name, value):
        """Generic (quick) search."""
        return queryset.filter(
            Q(device__name__icontains=value)
            | Q(destination__icontains=value)
            | Q(protocol__icontains=value)
            | Q(nexthop_ip__icontains=value)
            | Q(nexthop_if__name__icontains=value)
            | Q(vrf__name__icontains=value)
        )



"""Script used to check NetDoc integrity."""  # pylint: disable=invalid-name

__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from extras.reports import Report

from dcim.models import Interface
from ipam.models import Prefix, IPAddress

from netdoc import utils
from netdoc.models import ArpTableEntry


class InterfaceLabels(Report):
    """Check current Interface label compared to Interface name."""

    description = "Check Interface label compared to Interface name"

    def test_label(self):
        """Test Interface name/label integrity."""
        interface_qs = Interface.objects.all()
        for interface_o in interface_qs:
            name = interface_o.name
            label = interface_o.label

            if utils.normalize_interface_label(name) == label:
                self.log_success(interface_o)
            else:
                self.log_failure(
                    interface_o, "Interface label does not match interface name"
                )


class VRFIpPrefixIntegrityCheck(Report):
    """Check integrity between VRF, Interface, IPAddress, and Prefix.

    The report verifies that:
    * Interface.VRF == Interface.IPAddress.VRF -> warning if not
    * Interface.IPAddress.VRF in Prefix.VRF -> error if not

    If Interface.VRF != Interface.IPAddress.VRF the user should check if this is expected
    (e.g. when Interface is configured in global but the IPAddress is part of a VRF).

    If Prefix.VRF does not exist for Interface.IPAddress.VRF the user should create it.
    """

    description = (
        "Check (1) VRF between Interface and IP address,  and"
        + "(2) prefix exist for the IP address within the same VRF"
    )

    def test_vrf(self):
        """Test VRF between Interface and Interface's IPAddress."""
        interface_qs = Interface.objects.filter(ip_addresses__isnull=False)
        for interface_o in interface_qs:
            interface_vrf_o = interface_o.vrf
            ipaddress_vrf_o = interface_o.ip_addresses.first().vrf

            if interface_vrf_o == ipaddress_vrf_o:
                self.log_success(interface_o)
            else:
                self.log_warning(
                    interface_o,
                    f"VRF differs between interface ({interface_vrf_o}) "
                    + "and interface's IP address ({ipaddress_vrf_o})",
                )

    def test_interface_prefix(self):
        """Test Interface's IPAddress owned by a Prefix."""
        interface_qs = Interface.objects.filter(ip_addresses__isnull=False)
        for interface_o in interface_qs:
            for ip_address_o in interface_o.ip_addresses.all():
                vrf_o = ip_address_o.vrf
                network = (
                    f"{str(ip_address_o.address.network)}/"
                    + f"{str(ip_address_o.address.prefixlen)}"
                )
                prefix_qs = Prefix.objects.filter(prefix=network, vrf=vrf_o)
                if prefix_qs:
                    # Prefix found
                    self.log_success(interface_o)
                else:
                    # Prefix missing
                    self.log_failure(
                        interface_o,
                        "Prefix missing for IP address "
                        + f"{str(ip_address_o.address)} with VRF {vrf_o}",
                    )

    def test_address_prefix(self):
        """Test IPAddress owned by a Prefix."""
        ipaddress_qs = IPAddress.objects.all()
        for ip_address_o in ipaddress_qs:
            vrf_o = ip_address_o.vrf
            network = f"{str(ip_address_o.address.network)}/{str(ip_address_o.address.prefixlen)}"
            prefix_qs = Prefix.objects.filter(prefix=network, vrf=vrf_o)
            if prefix_qs:
                # Prefix found
                self.log_success(ip_address_o)
            else:
                # Prefix missing
                self.log_failure(
                    ip_address_o,
                    "Prefix missing for IP address "
                    + f"{str(ip_address_o.address)} with VRF {vrf_o}",
                )


class IPAMFromARP(Report):
    """Check corrispondence between ARP table and IPAM."""

    description = "Check corrispondence between ARP table and IPAM"

    def test_ipam(self):
        """For each ARP, check if an IPAddress exist.

        ARPTableEntry.Interface.IPAddress.VRF must be equal to IPAddress.VRF.
        """
        for arptableentry_o in ArpTableEntry.objects.all():
            try:
                interface_vrf_o = arptableentry_o.interface.ip_addresses.first().vrf
            except AttributeError:
                self.log_failure(
                    arptableentry_o,
                    f"IP address not found on interface {arptableentry_o.interface}, "
                    + "maybe some ingestion script has failed",
                )
                continue

            # IP address with prefixlen built from ARP table and associated interface
            address = str(arptableentry_o.ip_address.ip)
            prefixlen = (
                arptableentry_o.interface.ip_addresses.filter(
                    address__net_contains_or_equals=address
                )
                .first()
                .address.prefixlen
            )
            ip_address = f"{address}/{prefixlen}"

            # Query for the IP address in the IPAM
            ipaddresses_qs = IPAddress.objects.filter(
                vrf=interface_vrf_o, address=ip_address
            )
            if ipaddresses_qs:
                self.log_success(arptableentry_o)
            else:
                self.log_warning(
                    arptableentry_o,
                    f"IP address {str(arptableentry_o.ip_address)} not found in IPAM "
                    + f"for VRF {interface_vrf_o}",
                )



"""Ingestor for netmiko_hp_comware_display_ip_vpn_instance_instance_name."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface, vrf
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/hp_comware/display_ip_vpn-instance_instance-name # pylint: disable=line-too-long
        interfaces = item.get("interfaces")
        vrf_name = item.get("name") if item.get("name") else None

        # Get or create VRF
        vrf_o = None
        if vrf_name:
            vrf_o = vrf.get(name=vrf_name)
            if not vrf_o:
                vrf_data = {
                    "name": vrf_name,
                }
                vrf_o = vrf.create(mandatory_rd=False, **vrf_data)

        for intf in interfaces:
            interface_name = intf
            label = utils.normalize_interface_label(interface_name)

            # Get or create Interface
            interface_o = interface.get(device_id=device_o.id, label=label)
            if not interface_o:
                interface_data = {
                    "name": label,
                    "device_id": device_o.id,
                }
                interface_o = interface.create(**interface_data)

            # Update Interface
            data = {
                "vrf_id": vrf_o.id if vrf_o else None,
            }
            interface.update(interface_o, **data)

    # Update the log
    log.ingested = True
    log.save()



# Generated by Django 4.0.4 on 2022-06-26 15:05

import dcim.fields
import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import ipam.fields
import taggit.managers


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("extras", "0073_journalentry_tags_custom_fields"),
        ("dcim", "0153_created_datetimefield"),
        ("ipam", "0057_created_datetimefield"),
    ]

    operations = [
        migrations.CreateModel(
            name="Credential",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("username", models.CharField(blank=True, max_length=100)),
                ("password", models.CharField(blank=True, max_length=100)),
                ("enable_password", models.CharField(blank=True, max_length=100)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "Credential",
                "verbose_name_plural": "Credentials",
                "ordering": ("name",),
                "unique_together": {("name",)},
            },
        ),
        migrations.CreateModel(
            name="Discoverable",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("address", models.GenericIPAddressField()),
                ("mode", models.CharField(max_length=30)),
                ("discoverable", models.BooleanField(default=False)),
                (
                    "last_discovered_at",
                    models.DateTimeField(blank=True, editable=False, null=True),
                ),
                (
                    "credential",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="discoverables",
                        to="netdoc.credential",
                    ),
                ),
                (
                    "device",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="dcim.device",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="dcim.site",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "Device",
                "verbose_name_plural": "Devices",
                "ordering": ("mode", "address"),
                "unique_together": {("address", "mode")},
            },
        ),
        migrations.CreateModel(
            name="DiscoveryLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("command", models.CharField(editable=False, max_length=255)),
                ("configuration", models.BooleanField(default=False, editable=False)),
                ("parsed_output", models.JSONField(default=list, editable=False)),
                (
                    "raw_output",
                    models.TextField(blank=True, default="", editable=False),
                ),
                ("request", models.CharField(editable=False, max_length=255)),
                ("success", models.BooleanField(default=False, editable=False)),
                ("parsed", models.BooleanField(default=False, editable=False)),
                ("ingested", models.BooleanField(default=False, editable=False)),
                (
                    "discoverable",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="discoverylogs",
                        to="netdoc.discoverable",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "Log",
                "verbose_name_plural": "Logs",
                "ordering": ("created",),
            },
        ),
        migrations.CreateModel(
            name="RouteTableEntry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("destination", ipam.fields.IPAddressField(editable=False)),
                ("distance", models.IntegerField(editable=False)),
                ("metric", models.IntegerField(editable=False)),
                (
                    "nexthop_ip",
                    ipam.fields.IPAddressField(blank=True, editable=False, null=True),
                ),
                ("type", models.CharField(editable=False, max_length=30)),
                (
                    "device",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="dcim.device",
                    ),
                ),
                (
                    "nexthop_if",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="dcim.interface",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
                (
                    "vrf",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="ipam.vrf",
                    ),
                ),
            ],
            options={
                "verbose_name": "Route ",
                "verbose_name_plural": "Routes",
                "ordering": ("device", "type", "metric"),
                "unique_together": {
                    (
                        "device",
                        "destination",
                        "distance",
                        "metric",
                        "nexthop_if",
                        "nexthop_ip",
                        "type",
                    )
                },
            },
        ),
        migrations.CreateModel(
            name="MacAddressTableEntry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("mac_address", dcim.fields.MACAddressField(editable=False)),
                (
                    "vendor",
                    models.CharField(
                        blank=True, editable=False, max_length=255, null=True
                    ),
                ),
                ("vvid", models.IntegerField()),
                (
                    "interface",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="dcim.interface",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "MAC Address table entry",
                "verbose_name_plural": "MAC Address table entries",
                "ordering": ("mac_address", "vvid"),
                "unique_together": {("interface", "mac_address", "vvid")},
            },
        ),
        migrations.CreateModel(
            name="ArpTableEntry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                    ),
                ),
                ("ip_address", ipam.fields.IPAddressField(editable=False)),
                ("mac_address", dcim.fields.MACAddressField(editable=False)),
                (
                    "vendor",
                    models.CharField(
                        blank=True, editable=False, max_length=255, null=True
                    ),
                ),
                (
                    "interface",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="dcim.interface",
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
            ],
            options={
                "verbose_name": "ARP table entry",
                "verbose_name_plural": "ARP table entries",
                "ordering": ("ip_address",),
                "unique_together": {("interface", "ip_address", "mac_address")},
            },
        ),
    ]



"""Schema validation for DiscoveryLog."""

__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from netdoc.models import DiscoveryLog, Discoverable
from netdoc import utils


def get_schema():
    """Return the JSON schema to validate DiscoveryLog data."""
    return {
        "type": "object",
        "properties": {
            "command": {
                "type": "string",
            },
            "configuration": {
                "type": "boolean",
            },
            "details": {
                "type": "object",
                "properties": {
                    "command": {
                        "type": "string",
                    },
                    "enable": {
                        "type": "boolean",
                    },
                    "order": {
                        "type": "integer",
                    },
                    "template": {
                        "type": "string",
                    },
                    "supported": {
                        "type": "boolean",
                    },
                },
                "required": [
                    "command",
                    "enable",
                    "order",
                    "template",
                ],
            },
            "discoverable_id": {
                "type": "integer",
                "enum": list(Discoverable.objects.all().values_list("id", flat=True)),
            },
            "parsed_output": {
                "type": ["object", "array", "string"],
            },
            "raw_output": {
                "type": "string",
            },
            "template": {
                "type": "string",
            },
            "success": {
                "type": "boolean",
            },
            "parsed": {
                "type": "boolean",
            },
            "ingested": {
                "type": "boolean",
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new DiscoveryLog objects."""
    schema = get_schema()
    schema["required"] = [
        "command",
        "details",
        "discoverable_id",
        "raw_output",
        "template",
    ]
    return schema


def create(**kwargs):
    """Create a DiscoveryLog."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(DiscoveryLog, **kwargs)
    return obj


def get_list(**kwargs):
    """Get a list of DiscoveryLog objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(DiscoveryLog, **kwargs)
    return result


def update(obj, **kwargs):
    """Update a DiscoveryLog."""
    update_always = ["parsed_output", "parsed", "ingested"]

    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    kwargs_always = utils.filter_keys(kwargs, update_always)
    obj = utils.object_update(obj, **kwargs_always, force=True)
    return obj



"""Tasks executed via Netbox scripts or manually."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import importlib
import os
import logging
import pprint
import tempfile

from nornir.core.plugins.inventory import InventoryPluginRegister
from nornir import InitNornir
from nornir.core.filter import F

from django.conf import settings

from netdoc.nornir_inventory import AssetInventory
from netdoc.models import DiscoveryModeChoices
from netdoc import utils

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("netdoc", {})


def discovery(addresses=None, script_handler=None):
    """Discovery all or a list of IP addresses."""
    if not addresses:
        addresses = []
    ntc_template_dir = os.environ.get("NET_TEXTFSM")

    if not ntc_template_dir:
        raise ValueError("NET_TEXTFSM not set in configuration.py")
    if not os.path.exists(ntc_template_dir):
        raise ValueError(f"{ntc_template_dir} not found")

    # Configuring Nornir
    logger = logging.getLogger("nornir")
    logger.setLevel(logging.DEBUG)
    if script_handler:
        log_filename = os.path.join(
            tempfile.gettempdir(), f"nornir-{utils.get_random_string(6)}.log"
        )
    else:
        log_filename = PLUGIN_SETTINGS.get("NORNIR_LOG")
    file_h = logging.FileHandler(log_filename)
    file_h.setLevel(logging.DEBUG)
    logger.addHandler(file_h)

    # Load Nornir custom inventory
    InventoryPluginRegister.register("asset-inventory", AssetInventory)

    # Create Nornir inventory
    nrni = InitNornir(
        runner={
            "plugin": "threaded",
            "options": {
                "num_workers": 100,
            },
        },
        inventory={"plugin": "asset-inventory"},
        logging={"enabled": False},
    )

    if addresses:
        # Execute on a selected hosts only
        # See https://theworldsgonemad.net/2021/nornir-inventory/
        pprint.pprint(addresses)
        nrni = nrni.filter(F(hostname__in=addresses))

    # Starting discovery job
    nornir_addresses = nrni.dict().get("inventory").get("hosts").keys()
    pprint.pprint(nornir_addresses)
    if script_handler:
        if not nornir_addresses:
            script_handler.log_failure("Nornir inventory is empty")
            return ""
        script_handler.log_info(
            f"Norninr inventory includes {', '.join(nornir_addresses)}"
        )

    # Run discovery scripts
    for mode, description in DiscoveryModeChoices():
        # framework = mode.split("_").pop(0)
        platform = "_".join(mode.split("_")[1:])
        filtered_devices = nrni.filter(platform=platform)
        filtered_addresses = (
            filtered_devices.dict().get("inventory").get("hosts").keys()
        )
        if not filtered_addresses:
            if script_handler:
                script_handler.log_warning(f"No {description} device found")
            continue
        if script_handler:
            script_handler.log_info(f"Starting discovery of {description} devices")
            script_handler.log_info(
                f"Norninr inventory includes {', '.join(filtered_addresses)}"
            )
        # Call the discovery script
        try:
            module = importlib.import_module(f"netdoc.discoverers.{mode}")
        except ModuleNotFoundError as exc:
            raise ModuleNotFoundError(f"Discovery script not found for {mode}") from exc
        module.discovery(filtered_devices)

    if script_handler:
        script_handler.log_info("Discovery completed")

    # Return log
    with open(log_filename, "r", encoding="utf-8") as fh_o:
        output = fh_o.read()
    if script_handler:
        # Delete log file
        os.remove(log_filename)
        # Spawan ingest script
        script_handler.log_info("Spawning ingest script")
        utils.spawn_script("Ingest")

    return output



# Generated by Django 4.1.6 on 2023-03-17 14:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("netdoc", "0005_diagram_include_global_vrf_alter_diagram_details"),
    ]

    operations = [
        migrations.AlterField(
            model_name="routetableentry",
            name="metric",
            field=models.BigIntegerField(blank=True, editable=False, null=True),
        ),
    ]



# Generated by Django 4.1.6 on 2023-03-23 11:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("netdoc", "0006_alter_routetableentry_metric"),
    ]

    operations = [
        migrations.AddField(
            model_name="discoverylog",
            name="supported",
            field=models.BooleanField(default=True, editable=False),
        ),
    ]



"""Ingestor for netmiko_cisco_nxos_show_interfaces."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface, vrf, device
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_nxos/show_ip_interface # pylint: disable=line-too-long
        interface_name = item.get("interface")
        label = utils.normalize_interface_label(interface_name)
        vrf_name = item.get("vrf_name")
        ip_list = [item.get("primary_ip_address")] + item.get("secondary_ip_address")
        mask_list = [item.get("primary_ip_subnet")] + item.get("secondary_ip_subnet")
        ip_addresses = [
            f"{ipaddr}/{mask_list[index].split('/').pop()}"
            for index, ipaddr in enumerate(ip_list)
        ]

        # Get or create VRF
        vrf_o = None
        if vrf_name and vrf_name != "default":
            vrf_o = vrf.get(name=vrf_name)
            if not vrf_o:
                data = {
                    "name": vrf_name,
                }
                vrf_o = vrf.create(**data)

        # Get or create Interface
        interface_o = interface.get(device_id=device_o.id, label=label)
        if not interface_o:
            interface_data = {
                "name": label,
                "device_id": device_o.id,
            }
            interface_o = interface.create(**interface_data)

        # Update Interface
        data = {
            "vrf_id": vrf_o.id if vrf_o else None,
        }
        interface.update(interface_o, **data)
        interface.update_addresses(interface_o, ip_addresses=ip_addresses)

    # Update management IP address
    device.update_management(device_o, log.discoverable.address)

    # Update the log
    log.ingested = True
    log.save()



"""Ingestor for netmiko_cisco_ios_show_lldp_neighbors_detail."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface, device, discoverable, cable
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device
    neighbors_per_interface = utils.count_interface_neighbors(
        log.parsed_output, "local_interface"
    )

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_lldp_neighbors_detail # pylint: disable=line-too-long
        local_interface_name = item.get("local_interface")
        local_interface_label = utils.normalize_interface_label(local_interface_name)
        remote_management_ip = utils.normalize_ip_address_or_none(
            item.get("management_ip")
        )
        remote_name = utils.normalize_hostname(item.get("neighbor"))
        remote_interface_label = utils.get_remote_lldp_interface_label(
            port_id=item.get("neighbor_port_id"),
            port_description=item.get("neighbor_interface"),
            system_description=item.get("system_description"),
        )

        if not utils.is_hostname(remote_name):
            # Skip neighbors not announcing a valid hostname
            continue

        if neighbors_per_interface.get(local_interface_label) > 1:
            # Skip interfaces with multiple neighbors
            continue

        if utils.parent_interface(local_interface_name) or utils.parent_interface(
            remote_interface_label
        ):
            # Skip subinterfaces
            continue

        # Get or create local Interface
        local_interface_o = interface.get(
            device_id=device_o.id, label=local_interface_label
        )
        if not local_interface_o:
            local_interface_data = {
                "name": local_interface_name,
                "device_id": device_o.id,
            }
            local_interface_o = interface.create(**local_interface_data)

        # Get or create remote Device
        remote_device_o = device.get(remote_name)
        if not remote_device_o:
            remote_device_data = {
                "name": remote_name,
                "site_id": device_o.site.id,
            }
            remote_device_o = device.create(**remote_device_data)

        # Get or create remote Interface
        remote_interface_o = interface.get(
            device_id=remote_device_o.id, label=remote_interface_label
        )
        if not remote_interface_o:
            remote_interface_data = {
                "name": remote_interface_label,
                "device_id": remote_device_o.id,
            }
            remote_interface_o = interface.create(**remote_interface_data)

        # Create remote Discoverable if not exist
        if remote_management_ip:
            # List by device name
            remote_discoverable_list = discoverable.get_list(device__name=remote_name)
            # Add list by IP address
            remote_discoverable_list += discoverable.get_list(
                address=remote_management_ip
            )
            if not remote_discoverable_list:
                remote_discoverable_data = {
                    # Not adding devce_id: it will be added during ingestion
                    "site_id": log.discoverable.site.id,
                    "mode": log.discoverable.mode,
                    "address": remote_management_ip,
                    "credential_id": log.discoverable.credential.id,
                }
                discoverable.create(**remote_discoverable_data)

        # Link
        cable.link(
            left_interface_id=local_interface_o.id,
            right_interface_id=remote_interface_o.id,
        )

    # Update the log
    log.ingested = True
    log.save()



"""Ingestor for netmiko_cisco_ios_show_mac_address_table."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface, macaddresstableentry
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_mac-address-table # pylint: disable=line-too-long
        if not item.get("destination_port"):
            # Skip entries without associated interfaces
            continue
        interface_name = item.get("destination_port")[0]
        label = utils.normalize_interface_label(interface_name)
        vlan_id = int(item.get("vlan"))
        mac_address = utils.normalize_mac_address(item.get("destination_address"))

        interface_o = interface.get(device_id=device_o.id, label=label)
        if not interface_o:
            interface_data = {
                "name": label,
                "device_id": device_o.id,
            }
            interface_o = interface.create(**interface_data)

        macaddresstableentry_o = macaddresstableentry.get(
            interface_id=interface_o.id, vvid=vlan_id, mac_address=mac_address
        )
        if not macaddresstableentry_o:
            data = {
                "interface_id": interface_o.id,
                "vvid": vlan_id,
                "mac_address": mac_address,
            }
            macaddresstableentry.create(**data)

    # Update the log
    log.ingested = True
    log.save()



"""Schema validation for ArpTableEntry."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from ipam.models import IPAddress, VRF

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate IPAddress data."""
    return {
        "type": "object",
        "properties": {
            "address": {
                "type": "string",
            },
            "vrf_id": {
                "type": "integer",
                "enum": list(VRF.objects.all().values_list("id", flat=True)),
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new IPAddress objects."""
    schema = get_schema()
    schema["required"] = [
        "address",
    ]
    return schema


def create(**kwargs):
    """Create an IPAddress."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(IPAddress, **kwargs)
    return obj


def get_list(**kwargs):
    """Get a list of IPAddress objects."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(IPAddress, **kwargs)
    return result



"""Schema validation for Site."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from dcim.models import Site

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate Site data."""
    return {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
            },
            "description": {
                "type": "string",
            },
        },
    }


def get(name):
    """Return a Site."""
    obj = utils.object_get_or_none(Site, name=name)
    return obj


def get_list(**kwargs):
    """Get a list of Site objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Site, **kwargs)
    return result



"""Ingestor for netmiko_cisco_xr_hostname."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import re

from netdoc.schemas import device, discoverable
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    # See https://github.com/netbox-community/devicetype-library/tree/master/device-types
    vendor = "Cisco"
    name = log.parsed_output

    # Parsing hostname
    try:
        name = re.match(r".*hostname\ (\S+)$", name, re.MULTILINE | re.DOTALL).group(1)
    except AttributeError as exc:
        raise AttributeError(f"Failed to match HOSTNAME regex on {name}") from exc
    name = utils.normalize_hostname(name)

    # Get or create Device
    data = {
        "name": name,
        "site_id": log.discoverable.site.id,
        "manufacturer": vendor,
    }
    device_o = device.get(name=data.get("name"))
    if not device_o:
        device_o = device.create(**data)

    if not log.discoverable.device:
        # Link Device to Discoverable
        discoverable.update(log.discoverable, device_id=device_o.id)

    # Update the log
    log.ingested = True
    log.save()



"""Script used to manually import Discoverables."""  # pylint: disable=invalid-name

__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from datetime import date, timedelta
import re
import netaddr

from django.conf import settings

from dcim.models import Site, DeviceRole, Interface, Device
from ipam.models import IPAddress
from extras.scripts import (
    Script,
    ChoiceVar,
    ObjectVar,
    MultiObjectVar,
    TextVar,
    BooleanVar,
    IntegerVar,
)

from netdoc.models import (
    DiscoveryModeChoices,
    Discoverable as Discoverable_m,
    Credential as Credential_m,
    DiscoveryLog as DiscoveryLog_m,
    ArpTableEntry as ArpTableEntry_m,
    MacAddressTableEntry as MacAddressTableEntry_m,
    DeviceImageChoices,
)
from netdoc.utils import log_ingest, normalize_interface_label, normalize_hostname
from netdoc.schemas import discoverable
from netdoc.tasks import discovery

PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("netdoc", {})
NORNIR_LOG = PLUGIN_SETTINGS.get("NORNIR_LOG")


class CreateDeviceRole(Script):
    """Script used to create DeviceRole used in Diagram.

    DeviceRole.slug is used to draw the icon in diagrams.
    """

    class Meta:
        """Script metadata."""

        name = "Create device roles"
        description = "Add device roles based on available PNG."
        commit_default = True

    def run(self, data, commit):
        """Start the script."""
        icons = DeviceImageChoices()
        for key, value in icons:
            try:
                DeviceRole.objects.get(slug=key)
                self.log_info(f"DeviceRole {key} found")
            except DeviceRole.DoesNotExist:  # pylint: disable=no-member
                self.log_warning(f"Created DeviceRole {key}")
                DeviceRole.objects.create(name=value, slug=key)


class AddDiscoverable(Script):
    """Script used to generate AddDiscoverable."""

    class Meta:
        """Script metadata."""

        name = "Add and discover"
        description = "Add comma separated IP addresses and discover them."
        commit_default = True

    # Credentials
    credential = ObjectVar(
        model=Credential_m,
        description="Credential used to discover.",
        required=False,
    )

    # Discovery mode
    mode = ChoiceVar(
        choices=DiscoveryModeChoices.CHOICES,
        description="Discovery mode",
        required=True,
    )

    # Site
    site = ObjectVar(
        model=Site,
        description="Site associated with discovered devices",
        required=True,
    )

    # IP addresses to be discovered
    ip_addresses = TextVar(
        description="IP addresses separated by comma or space",
        required=True,
    )

    def run(self, data, commit):
        """Start the script."""
        discoverable_ip_addresses = []

        if not commit:
            self.log_warning("Commit not set, using dry-run mode")

        credential_o = data.get("credential")
        mode = data.get("mode")
        site_o = data.get("site")
        ip_addresses = re.split(" |,|\n", data.get("ip_addresses"))

        # Parse IP addresses
        for ip_address in ip_addresses:
            ip_address = ip_address.strip()
            if not ip_address:
                # Skip empty string
                continue

            try:
                netaddr.IPAddress(ip_address)
            except netaddr.core.AddrFormatError:
                # Skip invalid IP address
                self.log_warning(f"Skipping invalid IP address {ip_address}")
                continue

            # Create or get Discoverable
            discoverable_o, created = discoverable.get_or_create(
                address=ip_address,
                site_id=site_o.pk,
                mode=mode,
                credential_id=credential_o.pk,
                discoverable=True,
            )
            if created:
                self.log_info(
                    f"Created new discoverable with IP address {discoverable_o.address}"
                )
            else:
                self.log_info(
                    f"Using existing discoverable with IP address {discoverable_o.address}"
                )

            discoverable_ip_addresses.append(discoverable_o.address)

        if not discoverable_ip_addresses:
            self.log_failure("No valid IP address to discover")
            return ""

        self.log_info(f"Starting discovery on {', '.join(discoverable_ip_addresses)}")
        output = discovery(discoverable_ip_addresses, script_handler=self)

        self.log_info("Discovery completed")

        return output


class Discover(Script):
    """Script used to start discovery."""

    class Meta:
        """Script metadata."""

        name = "Discover"
        description = "Start discovery on one, many or all discoverables."
        commit_default = True

    # Discoverable
    discoverables = MultiObjectVar(
        model=Discoverable_m,
        query_params={
            "discoverable": True
        },  # TODO: not working, discoverable=False are also showed (cosmetic).
        description="Devices to be discovered (leave empty to discover everything).",
        required=False,
    )

    # Ingested?
    undiscovered_only = BooleanVar(
        description="Undiscovered devices only (the above setting is ignored).",
        required=False,
        default=True,
    )

    def run(self, data, commit):
        """Start the script."""
        discoverable_ip_addresses = []
        discoverables = data.get("discoverables")

        # Filtering out discoverable=False is done at Nornir inventory level.

        if data.get("undiscovered_only"):
            # Get only undiscovered IP addresses
            discoverables = discoverable.get_list(last_discovered_at__isnull=True)
            for discoverable_o in discoverables:
                discoverable_ip_addresses.append(discoverable_o.address)

            self.log_info(
                f"Starting first discovery on {', '.join(discoverable_ip_addresses)}"
            )
            output = discovery(discoverable_ip_addresses, script_handler=self)
        elif discoverables:
            for discoverable_o in discoverables:
                discoverable_ip_addresses.append(discoverable_o.address)

            self.log_info(
                f"Starting discovery on {', '.join(discoverable_ip_addresses)}"
            )
            output = discovery(discoverable_ip_addresses, script_handler=self)
        else:
            self.log_info("Starting discovery on all IP addresses")
            output = discovery(None, script_handler=self)

        return output


class Ingest(Script):
    """Script used to start ingestion."""

    class Meta:
        """Script metadata."""

        name = "Ingest"
        description = (
            "Start data ingestion (automatically triggered after a discovery)."
        )
        commit_default = True

    # Discoverable
    discoverables = MultiObjectVar(
        model=Discoverable_m,
        description="Limit ingestion to selected discoverables.",
        required=False,
    )

    # Ingested?
    re_ingest = BooleanVar(
        description="Force re-ingestion.",
        required=False,
    )

    def run(self, data, commit):
        """Start the script."""
        log_queryset = DiscoveryLog_m.objects.filter(parsed=True).order_by("order")
        if not data.get("re_ingest"):
            # Filter out logs already ingested
            log_queryset = log_queryset.filter(ingested=False)
        if data.get("discoverables"):
            log_queryset = log_queryset.filter(
                discoverable__in=data.get("discoverables")
            )

        if log_queryset:
            self.log_info(f"Ingesting {len(log_queryset)} logs")
        else:
            self.log_failure("No log to ingest")

        for log in log_queryset:
            if log.ingested:
                self.log_info(
                    f"Reingesting log {log.id} with command {log.command} "
                    + f"on device {log.discoverable}"
                )
            else:
                self.log_info(
                    f"Ingesting log {log.id} with command {log.command} on "
                    + f"device {log.discoverable}"
                )
            log_ingest(log)


class Purge(Script):
    """Script used to delete old data."""

    class Meta:
        """Script metadata."""

        name = "Purge"
        description = "Delete old logs, ARP table entries, MAC Address table entries."
        commit_default = True

    # Minimum days to delete
    days = IntegerVar(
        description="Minimum age in days to delete",
        required=True,
        default=90,
    )

    def run(self, data, commit):
        """Start the script."""
        today = date.today()
        today_minus_x = today - timedelta(days=data.get("days"))

        discoverylogs_qs = DiscoveryLog_m.objects.filter(created__lt=today_minus_x)
        arpentries_qs = ArpTableEntry_m.objects.filter(last_updated__lt=today_minus_x)
        macaddressentries_qs = MacAddressTableEntry_m.objects.filter(
            last_updated__lt=today_minus_x
        )

        self.log_info(f"Deleting entries updated before f{today_minus_x}")
        self.log_info(f"Deleting {len(discoverylogs_qs)} discovery logs")
        if discoverylogs_qs:
            discoverylogs_qs.delete()
        self.log_info(f"Deleting {len(arpentries_qs)} ARP table entries")
        if arpentries_qs:
            arpentries_qs.delete()
        self.log_info(f"Deleting {len(macaddressentries_qs)} MAC address table entries")
        if macaddressentries_qs:
            macaddressentries_qs.delete()
        self.log_info("Purge completed")


class IPAMFromARP(Script):
    """Create IP Address on IPAM based on ARP table."""

    class Meta:
        """Script metadata."""

        name = "Update IPAM from ARP tables"
        description = (
            "Create IP Address on IPAM based on ARP table. Should be started after "
            + "VRFIpPrefixIntegrityCheck report."
        )
        commit_default = True

    def run(self, data, commit):
        """For each ARP, check if an IPAddress exist and create it if it doesn't.

        ARPTableEntry.Interface.IPAddress.VRF must be equal to Prefix.VRF.
        """
        for arptableentry_o in ArpTableEntry_m.objects.all():
            try:
                interface_vrf_o = arptableentry_o.interface.ip_addresses.first().vrf
            except AttributeError:
                self.log_failure(
                    f"IP address not found on interface {arptableentry_o.interface}, maybe "
                    + "some ingestion script has failed",
                )
                continue

            # IP address with prefixlen built from ARP table and associated interface
            address = str(arptableentry_o.ip_address.ip)
            prefixlen = (
                arptableentry_o.interface.ip_addresses.filter(
                    address__net_contains_or_equals=address
                )
                .first()
                .address.prefixlen
            )
            ip_address = f"{address}/{prefixlen}"

            # Query for the IP address in the IPAM
            ipaddresses_qs = IPAddress.objects.filter(
                vrf=interface_vrf_o, address=ip_address
            )
            if not ipaddresses_qs:
                IPAddress.objects.create(address=ip_address, vrf=interface_vrf_o)
                self.log_info(
                    f"IP address {ip_address} with VRF {interface_vrf_o} added in IPAM",
                )


class FixData(Script):
    """Fix Device names and Interface labels."""

    class Meta:
        """Script metadata."""

        name = "Fix Netbox data"
        description = (
            "Fix Netbox data to be used with NetDoc (Device names, Interface labels)."
        )
        commit_default = True

    def run(self, data, commit):
        """Test and fix Netbox data."""
        # Test Device name
        device_qs = Device.objects.all()
        for device_o in device_qs:
            name = device_o.name
            new_name = normalize_hostname(name)

            if name == new_name:
                self.log_info(f"{device_o} has name {new_name}")
            else:
                self.log_warning(f"Renaming device {device_o} to {new_name}")
                device_o.name = new_name
                device_o.save()

        # Test Interface name/label integrity
        interface_qs = Interface.objects.all()
        for interface_o in interface_qs:
            name = interface_o.name
            new_label = normalize_interface_label(name)

            if interface_o.label == new_label:
                self.log_info(f"{interface_o} has label {new_label}")
            else:
                self.log_warning(
                    f"Renaming label {interface_o.label} to {new_label} for interface {interface_o}"
                )
                interface_o.label = new_label
                interface_o.save()



"""Ingestor for netmiko_cisco_ios_show_interfaces_switchport."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import interface
from netdoc import utils


def ingest(log):
    """Processing parsed output."""
    device_o = log.discoverable.device

    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/cisco_ios/show_interfaces_switchport # pylint: disable=line-too-long
        interface_name = item.get("interface")
        label = utils.normalize_interface_label(interface_name)
        mode = utils.normalize_interface_mode(item.get("mode"))
        native_vlan = int(item.get("native_vlan")) if item.get("native_vlan") else None
        access_vlan = (
            int(item.get("access_vlan"))
            if item.get("access_vlan") and item.get("access_vlan") != "unassigned"
            else None
        )
        tagged_vlans = utils.normalize_vlan_list(item.get("trunking_vlans"))

        # Get or create Interface
        interface_o = interface.get(device_id=device_o.id, label=label)
        if not interface_o:
            interface_data = {
                "name": label,
                "device_id": device_o.id,
            }
            interface_o = interface.create(**interface_data)

        # Update Interface mode
        mode_data = {
            "untagged_vlan": access_vlan if mode == "access" else native_vlan,
            "mode": mode,
            "tagged_vlans": tagged_vlans,
        }
        interface.update_mode(interface_o, **mode_data)

    # Update the log
    log.ingested = True
    log.save()
